// danse.H  Version 7.4e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-09 08:42:25 tac>
//
// 各プログラムに共通するヘッダファイルや定数、自作関数のプロトタイプ宣言を行う。
//
// 関連ライブラリ:libdanse.so.7.4

#ifndef _DANS                   // 重複インクルードを防ぐため
#define _DANS

#include <mate.H>               // 行列クラス用ヘッダファイル(fune.H含む)

namespace dans                  // dans用名前空間定義
{
  using namespace matvec; // 名前空間matvecを合成する(名前空間fun含む)

/* 定数定義 */
  const string Version = "[DANS V7.4e]"; // プログラムのバージョン
  constexpr char CHRdelm = ',';  // データの区切り文字
  constexpr char CHRtags = '<';  // タグ開始文字
  constexpr char CHRtage = '>';  // タグ終了文字
  constexpr char CHRetag = '/';  // 終了タグ識別文字
  constexpr char CHRhead = '^';  // 正規表現の先頭指定文字
  constexpr char CHRtail = '$';  // 正規表現の末尾指定文字
  constexpr char CHRstr = '/';   // 書式の文字列指定文字
  constexpr char CHRrep = 'x';   // 書式の繰り返し数指定文字
  const string BLANKs = "";           // 文字欠損データ
  constexpr double BLANKf = HUGE_VAL; // 数字欠損データ(doubleの無限大値)
  constexpr double FDmin = (double)FLT_MIN; // 数字データ最小値(floatの最小値)
  constexpr double FDmax = (double)FLT_MAX; // 数字データ最大値(floatの最大値)
  constexpr double ALdef = 0.05; // 省略時有意水準
  constexpr double CCdef = 0.95; // 省略時信頼係数
  constexpr double POdef = 0.8;  // 省略時検出力
  const string TAGtitle = "t";   // タイトル指定タグ
  const string TAGstra = "s";    // 層別内容指定タグ
  const string TAGitem = "i";    // 項目名指定タグ
  const string ATTtype = "type"; // 属性名
  const string AVALand = "and"; // 属性値:AND層別項目
  const string AVALor = "or";   // 属性値:OR層別項目
  const string AVALrow = "row"; // 属性値:縦(行)項目
  const string AVALcol = "col"; // 属性値:横(列)項目
  const string NGPdelm = "%__"; // Ngraphデータの区切り文字列

  enum datatype                 // データの種類
  {
    datanon,                    // なし
    dfdata,                     // 実数データ
    dsdata,                     // 文字データ
    dataetc                     // その他(エラー)
  };

  enum tagtype                  // タグの種類
  {
    tagnon,                     // なし
    ttitle,                     // タイトル
    tstra,                      // 層別情報
    titem,                      // 項目情報
    tagetc                      // その他(エラー)
  };

  enum atttype                  // 属性の種類
  {
    attnon,                     // なし
    atype,                      // タイプ
    attetc                      // その他(エラー)
  };

  enum avaltype                 // 属性値の種類
  {
    avalnon,                    // なし
    avaland,                    // AND層別項目
    avalor,                     // OR層別項目
    avalrow,                    // 縦(行)項目
    avalcol,                    // 横(列)項目
    avaletc                     // その他(エラー)
  };

  enum formtype                 // データ識別情報の出力形式
  {
    formnon,                    // 出力帳票形式
    formtag,                    // タグ形式
    formngp,                    // Ngraph形式
    formetc                     // その他(エラー)
  };

  enum tabtype                  // 度数表の種類
  {
    tabnon,                     // 出力なし
    tabcode,                    // コードだけ出力(freqスクリプト用)
    tabdata,                    // 度数データ形式
    tabfreq,                    // 度数表形式
    tabfprc,                    // 列合計に対する%付き度数表形式
    tabfprr,                    // 行合計に対する%付き度数表形式
    tabfprt,                    // 全例数に対する%付き度数表形式
    tabngp,                     // Ngraph用度数分布データ
    tabetc                      // その他
  };

  enum testtype                 // 検定の種類
  {
    testnon,                    // 検定なし
    testorg,                    // 連続修正無の検定
    testcor,                    // 連続修正有の検定
    testaut,                    // 連続修正自動判定の検定
    testetc                     // その他
  };

  enum mctype                   // 多重比較の種類
  {
    mcnon,                      // なし
    mcbon,                      // Bonferroni型
    mcdun,                      // Dunnet型
    mcfis,                      // Fisher型
    mcsch,                      // Scheffe型
    mctuk,                      // Tukey型
    mcetc                       // その他
  };

  enum mattype                  // 行列の種類
  {
    matnon,                     // なし
    mats,                       // 積和行列
    matv,                       // 分散行列
    matr,                       // 相関行列
    matetc                      // その他
  };

  enum octype                   // 転帰の種類
  {
    ocnon,                      // なし
    ocdeath,                    // 死亡
    ocalive,                    // 生存
    ocdropout,                  // 脱落
    ocetc                       // その他
  };

  enum outtype                  // 出力データの種類
  {
    outnon,                     // 出力なし
    outdns,                     // データファイル形式のデータ
    outngp,                     // Ngraph用データ
    outres,                     // 分析結果
    outetc                      // その他
  };

/* dans用データクラス定義(ソースファイル:datac.C) */
  class datac                   // ケースデータクラス
  {
  private:
    string blank;               // ダミーブランクデータ
    string vdstr;               // 特殊文字列表現データ
    size_t dstridx;             // 特殊文字列表現データのインデックス
    double vfdata;              // 実数データ
    size_t fdataidx;            // 実数データのインデックス
    vecstr vsdatavec;           // 文字列型データベクトル配列
  public:
    datac() : blank(BLANKs), vdstr(BLANKs), dstridx(UINT_MAX), vfdata(BLANKf), fdataidx(UINT_MAX), vsdatavec()
    // デフォールトコンストラクタ
    {}
    explicit datac(size_t n) : blank(BLANKs), vdstr(BLANKs), dstridx(UINT_MAX), vfdata(BLANKf), fdataidx(UINT_MAX), vsdatavec(n, BLANKs)
    // コンストラクタ
    {}
    explicit datac(const vecstr& strvec) : blank(BLANKs), vdstr(BLANKs), dstridx(UINT_MAX), vfdata(BLANKf), fdataidx(UINT_MAX), vsdatavec(strvec)
    // 文字列ベクトル配列による型変換コンストラクタ
    {}
    datac(const datac& data0) : blank(data0.blank), vdstr(data0.vdstr), dstridx(data0.dstridx), vfdata(data0.vfdata), fdataidx(data0.fdataidx), vsdatavec(data0.vsdatavec)
    // コピーコンストラクタ
    {}
    virtual ~datac()            // デストラクタ
    {}
    operator vecstr() const     // 文字列ベクトル配列型変換演算子
    { return vsdatavec; }
    const datac& operator=(const datac& data0); // コピー代入演算子
    const datac& operator=(const string& sdata0);
    // 文字データ代入演算子(全データ同値代入)
    const datac& operator=(double fdata0);
    // 実数データ代入演算子(全データ同値代入)
    bool empty() const          // データが空がどうかの判定
    { return vsdatavec.empty(); }
    size_t size() const         // データ数を返す
    { return vsdatavec.size(); }
    size_t n() const            // データ数を返す
    { return size(); }
    void clear();               // 全データを削除する
    void erase()                // clear()の別名
    { clear(); }
    int erase(size_t idxbeg, size_t idxend = UINT_MAX);
    // 指定された範囲のデータを削除する
    size_t resize(size_t n0);   // データ数を変更する
    size_t n(size_t n0)         // データ数を変更する
    { return resize(n0); }
    const vecstr& sdatavec() const // 文字列ベクトル配列への参照を返す
    { return vsdatavec; }
    const string& sdata(size_t idx) const     // 文字データを返す
    { return (idx < vsdatavec.size()) ? vsdatavec[idx] : blank; }
    const string& assign(size_t idx, const string& sdata0);
    // 文字データを設定する
    const string& sdata(size_t idx, const string& sdata0)
    // 文字データを設定する(省略形)
    { return assign(idx, sdata0); }
    const string& dstr(size_t idx);
    // 特殊文字列を用いた表示用データを返す
    double fdata(size_t idx);   // 実数データを返す
    double assign(size_t idx, double fdata0); // 実数データを設定する
    double fdata(size_t idx, double fdata0) // 実数データを設定する(省略形)
    { return assign(idx, fdata0); }
    void push_back(const string& sdata0) // 文字データを末尾に追加する
    { vsdatavec.push_back(sdata0); }
    int insert(size_t idx, const string& sdata0); // 文字データを挿入する
    int put(ostream& fout);     // ケースデータをファイルに出力する
    int getdata(const lstr& strlst);
    // 文字列リストからデータを取り出し、取り出したデータ数を返す
    int getdata(const string& str, char delmc = CHRdelm);
    // 文字列からデータを取り出し、取り出したデータ数を返す
  };

  using ldata = list<datac>;    // データリストの短名定義
  using ldata_itr = list<datac>::iterator; // 反復子の短名定義
  using ldata_citr = list<datac>::const_iterator; // 反復子の短名定義

/* dans用層別クラス定義(ソースファイル:stra.C) */
  class stra                    // 層別情報クラス
  {
  private:
    string vstrastr;            // 層別情報の文字列表現
    datatype vstratype;         // 層別項目の種別
    size_t vstraino;            // 層別項目番号
    pairdbl vrange;             // 実数データの範囲
    lstr vstraslst;             // 文字データリスト
  public:
    stra() : vstrastr(), vstratype(datanon), vstraino(0), vrange(BLANKf, BLANKf), vstraslst()
    // デフォールトコンストラクタ
    {}
    stra(int straino0, const lstr& straslst0) : vstrastr(), vstratype(dsdata), vstraino(straino0), vrange(BLANKf, BLANKf), vstraslst(straslst0)
    // 文字データ項目によるコンストラクタ
    {}
    stra(int straino0, pairdbl range0) : vstrastr(), vstratype(dfdata), vstraino(straino0), vrange(range0), vstraslst()
    // 実数データ項目によるコンストラクタ
    {}
    stra(const stra& stra0) : vstrastr(stra0.vstrastr), vstratype(stra0.vstratype), vstraino(stra0.vstraino), vrange(stra0.vrange), vstraslst(stra0.vstraslst)
    // コピーコンストラクタ
    {}
    virtual ~stra()             // デストラクタ
    {}
    const stra& operator=(const stra& stra0); // コピー代入演算子
    void clear();                             // 層別情報を削除する
    const string& strastr(const string& iname);
    // 層別情報の文字列表現を返す
    datatype stratype() const   // 層別項目の種別を返す
    { return vstratype; }
    size_t straino() const      // 層別項目番号を返す
    { return vstraino; }
    datatype get(char strach, size_t straino0, const string& str);
    // 層別情報を取得する
    bool match(datac& cdata) const; // 層別条件に当てはまるかどうか判定する
  };

  using lstra = list<stra>;               // 層別リストの短名定義
  using lstra_itr = list<stra>::iterator; // 反復子の短名定義
  using lstra_citr = list<stra>::const_iterator; // 反復子の短名定義

  extern bool straok(const lstra& stralst, size_t ni);
  // 層別情報の整合性をチェックする
  extern bool stramatch(const lstra& stralst, datac& cdata);
  // 層別に当てはまるかどうか判定する

/* dans用データ識別情報クラス定義(ソースファイル:idd.C) */
  class element                 // 要素クラス
  {
  private:
    tagtype vtag;               // タグの種類
    string vpcdata;             // 内容
    atttype vatt;               // 属性の種類
    avaltype vaval;             // 属性値
    string vestr;               // 要素の文字列表現
    tagtype search_tag(const string& tname) const;
    // タグ名からタグの種類をサーチして返す
    atttype search_att(const string& aname) const;
    // 属性名から属性の種類をサーチして返す
    avaltype search_aval(const string& avname) const;
    // 属性値名から属性値の種類をサーチして返す
  public:
    element() : vtag(tagnon), vpcdata(), vatt(attnon), vaval(avalnon), vestr()
    // デフォールトコンストラクタ
    {}
    element(tagtype tag0, const string& pcdata0, atttype att0 = attnon, avaltype aval0 = avalnon) : vtag(tag0), vpcdata(pcdata0), vatt(att0), vaval(aval0), vestr()
    // コンストラクタ
    {}
    element(const element& element0) : vtag(element0.vtag), vpcdata(element0.vpcdata), vatt(element0.vatt), vaval(element0.vaval), vestr(element0.vestr)
    // コピーコンストラクタ
    {}
    virtual ~element()          // デストラクタ
    {}
    const element& operator=(const element& element0); // コピー代入演算子
    void clear();               // 要素を削除する
    tagtype tag() const         // タグの種類を返す
    { return vtag; }
    tagtype tag(tagtype tag0);  // タグの種類を設定する
    const string& pcdata() const // 内容を返す
    { return vpcdata; }
    const string& pcdata(const string& pcdata0); // 内容を設定する
    atttype att() const                          // 属性の種類を返す
    { return vatt; }
    atttype att(atttype att0);  // 属性の種類を設定する
    avaltype aval() const       // 属性値を返す
    { return vaval; }
    avaltype aval(avaltype aval0); // 属性値を設定する
    const string& assign(tagtype tag0, const string& pcdata0, atttype att0 = attnon, avaltype aval0 = avalnon);
    // 要素を設定する
    const string& estr();       // 要素の文字列表現を返す
    tagtype get(const string& str); // 文字列から要素を取得する
  };

  class idd                     // データ識別情報クラス
  {
  private:
    element vetitle;            // タイトル要素
    vector<element> vestravec;  // 層別情報ベクトル配列
    vector<element> veitemvec;  // 項目情報ベクトル配列
    element edummy;             // ダミー要素
    size_t vnc;                 // ケース数
    string iname;               // ダミー項目名用
  public:
    idd() : vetitle(), vestravec(), veitemvec(), edummy(), vnc(0), iname()
    // デフォールトコンストラクタ
    {}
   idd(const string& title, size_t ni0) : vetitle(ttitle, title), vestravec(), veitemvec(ni0), edummy(), vnc(0), iname()
    // コンストラクタ
    {}
    idd(const idd& idd0) : vetitle(idd0.vetitle), vestravec(idd0.vestravec), veitemvec(idd0.veitemvec), edummy(idd0.edummy), vnc(idd0.vnc), iname(idd0.iname)
    // コピーコンストラクタ
    {}
    virtual ~idd()              // デストラクタ
    {}
    const idd& operator=(const idd& idd0); // コピー代入演算子
    void clear();                          // データ識別情報を削除する
    size_t set(const string& title, size_t ni0); // データ識別情報を設定する
    const string& title() const                  // データ名を返す
    { return vetitle.pcdata(); }
    size_t ns() const           // 層別数を返す
    { return vestravec.size(); }
    element& estra(size_t idx)  // 層別情報要素を返す
    { return (idx < vestravec.size()) ? vestravec[idx] : edummy; }
    size_t ni() const           // 項目数を返す
    { return veitemvec.size(); }
    element& eitem(size_t idx)  // 項目情報要素を返す
    { return (idx < veitemvec.size()) ? veitemvec[idx] : edummy; }
    const string& assign_item(size_t idx, const string& pcdata0, atttype att0 = attetc, avaltype aval0 = avaletc);
    // 項目情報を設定する
    const string& item(size_t idx) const // 項目名を返す
    { return (idx < veitemvec.size()) ? veitemvec[idx].pcdata() : iname; }
    const string& item(size_t idx, const string& item0) // 項目名を設定する
    { return assign_item(idx, item0); }
    size_t nc() const           // ケース数を返す
    { return vnc; }
    size_t nc(size_t nc0)       // ケース数を設定する
    { return (vnc = nc0); }
    int put(ostream& fout, formtype formt = formnon);
    // データ識別情報をファイルに出力する
    int put_title(ostream& fout, formtype formt = formnon);
    // データ名をファイルに出力する
    int put_stra(ostream& fout, formtype formt = formnon);
    // 層別情報をファイルに出力する
    int put_item(ostream& fout, formtype formt = formnon);
    // 項目情報をファイルに出力する
    int put_item(ostream& fout, size_t idx, formtype formt = formnon);
    // 項目をファイルに出力する
    int append(const string& str);
    // データ識別情報を追加する
    int append_stra(const string& pcdata0, atttype att0, avaltype aval0);
    // 層別情報を追加する
    int append_stra(lstra& stralst);
    // 多重層別情報を追加する
    int append_item(const string& pcdata0, atttype att0, avaltype aval0);
    // 項目情報を追加する
  };

  using listidd = list<idd>;      // データ識別情報リストの別名定義

/* 書式クラス定義(ソースファイル:format.C) */
  class format                  // 書式クラス
  {
  private:
    pairint vnf;                // データの桁数
    string vstr;                // 文字列
  public:
    format() : vnf(0, -1), vstr() // デフォールトコンストラクタ
    {}
    format(int n0, int f0, const string& str0) : vnf(n0, f0), vstr(str0)
    // コンストラクタ
    {}
    format(const format& format0) : vnf(format0.vnf), vstr(format0.vstr)
    // コピーコンストラクタ
    {}
    virtual ~format()           // デストラクタ
    {}
    const format& operator=(const format& format0); // コピー代入演算子
    void clear();                                   // 書式を削除する
    int n() const               // データの全桁数を返す
    { return vnf.first; }
    int n(int n0)               // データの全桁数を設定する
    { return (vnf.first = n0); }
    int f() const               // データの小数部桁数を返す
    { return vnf.second; }
    int f(int f0)               // データの小数部桁数を設定する
    { return (vnf.second = f0); }
    const string& str() const   // 文字列を返す
    { return vstr; }
    const string& str(const string& str0) // 文字列を設定する
    { return (vstr = str0); }
    int getfmt(const string& str); // 文字列から書式を取り出す
  };

  int getfmtlst(list<format>& fmtlst, const string& sstr, char delmc);
// 文字列から書式を読み込み書式リストに追加する

/* 基礎統計量クラス定義(クラステンプレート) */
  template <class T> class statv // 基礎統計量ベースクラス
  {
  protected:                    // 変数をサブクラスに継承するため
    int vn;                     // 例数
    T sigmax;                   // Σx
    T sigmax2;                  // Σx^2
  public:
    statv() : vn(0), sigmax((T)0.0), sigmax2((T)0.0)
    // デフォールトコンストラクタ
    {}
    explicit statv(T x) : vn(1), sigmax(x), sigmax2(x * x)
    // データによるコンストラクタ
    {}
    statv(int n0, T sigamax0, T sigmax20) : vn(n0), sigmax(sigamax0), sigmax2(sigmax20)
    // 例数等によるコンストラクタ
    {}
    statv(const statv<T>& statv0) : vn(statv0.vn), sigmax(statv0.sigmax), sigmax2(statv0.sigmax2)
    // コピーコンストラクタ
    {}
    virtual ~statv()            // デストラクタ
    {}
    const statv<T>& operator=(const statv<T>& statv0) // コピー代入演算子
    {
      if (&statv0 == this) return *this; // 自分自身への代入を防ぐ
      vn = statv0.vn;  sigmax = statv0.sigmax;  sigmax2 = statv0.sigmax2;
      return *this;             // 自分自身への参照を返す
    }
    const statv<T>& operator+=(const statv<T>& statv0) // 加算代入演算子
    {
      if (statv0.vn < 1) return *this; // 加算しない
      vn += statv0.vn;  sigmax = sigmax + statv0.sigmax;
      sigmax2 = sigmax2 + statv0.sigmax2;
      return *this;                 // 自分自身への参照を返す
    }
    virtual void clear()        // データをクリアする
    { vn = 0;  sigmax = sigmax2 = (T)0.0; }
    virtual int set(T x)        // データによる設定
    {
      if (x == (T)BLANKf) clear();
      else { vn = 1;  sigmax = x;  sigmax2 = x * x; }
      return vn;
    }
    int set(int n0, T mean0, T sd0) // 例数、平均、標準偏差による設定
    {
      if (n0 < 1)
      { clear(); }
      else
      {
        vn = n0;
        sigmax = (T)n0 * mean0;
        sigmax2 = (T)(n0 - 1) * sd0 * sd0 + sigmax * mean0;
      }
      return vn;
    }
    virtual int sigma(T x)   // データを累積する(サブクラスで変更可能)
    {
      if (x == (T)BLANKf) return 0; // 累積せず
      ++vn;  sigmax = sigmax + x;  sigmax2 = sigmax2 + x * x;
      return 1;                     // 累積
    }
    int n() const                // 例数を返す
    { return vn; }
    T sum() const               // 合計を返す
    { return sigmax; }
    T mean() const              // 平均を返す
    { return (vn > 0) ? sigmax / (T)vn : (T)0.0; }
    T sum2() const              // 平方合計を返す
    { return sigmax2; }
    T ss() const                // 平方和を返す
    {
      if (vn < 1) return (T)0.0;
      T sxx = sigmax2 - sigmax * sigmax / (T)vn;
      const T Seps = numeric_limits<T>::epsilon() * fmax((T)FDmin, sigmax2);
      return (sxx < Seps) ? (T)0.0 : sxx;
    }
    T vr() const                // 不偏分散を返す
    { return (vn > 1) ? ss() / (T)(vn - 1) : (T)0.0; }
    T vrs() const               // 標本分散を返す
    { return (vn > 1) ? ss() / (T)vn : (T)0.0; }
    T sd() const                // 不偏標準偏差を返す
    { return sqrt(vr()); }
    T sds() const               // 標本標準偏差を返す
    { return sqrt(vrs()); }
    T se() const                // 標準誤差を返す
    { return (vn > 1) ? sqrt(vr() / (T)vn) : (T)0.0; }
  };

  template <class T> class statk : public statv<T> // 基礎統計量+k統計量クラス
  {
  protected:                    // 変数をサブクラスに継承するため
    T sigmax3;                  // Σx^3
    T sigmax4;                  // Σx^4
  public:
    statk() : statv<T>(), sigmax3((T)0.0), sigmax4((T)0.0)
    // デフォールトコンストラクタ
    {}
    explicit statk(T x) : statv<T>(x), sigmax3(x * x * x), sigmax4(x * x * x * x)
    // データによるコンストラクタ
    {}
    statk(int n0, T sig0, T sig20, T sigmax30, T sigmax40) : statv<T>(n0, sig0, sig20), sigmax3(sigmax30), sigmax4(sigmax40)
    // 例数等によるコンストラクタ
    {}
    statk(const statk<T>& statk0) : statv<T>(statk0.n(), statk0.sum(), statk0.sum2()), sigmax3(statk0.sigmax3), sigmax4(statk0.sigmax4)
    // コピーコンストラクタ
    {}
    virtual ~statk()            // デストラクタ
    {}
    const statk<T>& operator=(const statk<T>& statk0) // コピー代入演算子
    {
      if (&statk0 == this) return *this; // 自分自身への代入を防ぐ
      this->vn = statk0.vn;
      this->sigmax = statk0.sigmax;  this->sigmax2 = statk0.sigmax2;
      this->sigmax3 = statk0.sigmax3;  this->sigmax4 = statk0.sigmax4;
      return *this;
    }
    const statk<T>& operator+=(const statk<T>& statk0) // 加算代入演算子
    {
      if (statk0.vn < 1) return *this; // 加算しない
      this->vn += statk0.vn;
      this->sigmax = this->sigmax + statk0.sigmax;
      this->sigmax2 = this->sigmax2 + statk0.sigmax2;
      this->sigmax3 = this->sigmax3 + statk0.sigmax3;
      this->sigmax4 = this->sigmax4 + statk0.sigmax4;
      return *this;             // 自分自身への参照を返す
    }
    virtual void clear()        // データをクリアする
    {
      this->vn = 0;
      this->sigmax = this->sigmax2 = this->sigmax3 = this->sigmax4 = (T)0.0;
    }
    virtual int sigma(T x)      // データを累積する
    {
      if (x == (T)BLANKf) return 0; // 累積せず
      T x2 = x * x;
      ++(this->vn);  this->sigmax = this->sigmax + x;
      this->sigmax2 = this->sigmax2 + x2;
      this->sigmax3 = this->sigmax3 + x2 * x;
      this->sigmax4 = this->sigmax4 + x2 * x2;
      return 1;                 // 累積
    }
    T m2() const                // m2統計量(標本分散)を返す
    { return this->vrs(); }
    T m3() const                // m3統計量を返す
    {
      if (this->vn < 3) return (T)0.0;
      ldbl s3 = this->sigmax3 - (T)3.0 * this->mean() * this->sigmax2 + (T)2.0 * this->mean() * this->mean() * this->sigmax;
      return s3 / (T)(this->vn);
    }
    T m4() const                // m4統計量を返す
    {
      if (this->vn < 4) return (T)0.0;
      T s4 = this->sigmax4 - (T)4.0 * this->mean() * this->sigmax3 + (T)6.0 * this->mean() * this->mean() * this->sigmax2 - (T)3.0 * this->mean() * this->mean() * this->mean() * this->sigmax;
      return s4 / (T)(this->vn);
    }
    T k2() const                // Fisherのk2統計量(不偏分散)を返す
    { return this->vr(); }
    T k3() const                // Fisherのk3統計量を返す
    {
      if (this->vn < 3) return (T)0.0;
      return m3() * (T)(this->vn) * (T)(this->vn) / (T)(this->vn - 1) / (T)(this->vn - 2);
    }
    ldbl k4() const             // Fisherのk4統計量を返す
    {
      if (this->vn < 4) return (T)0.0;
      T s4 = (T)(this->vn) * (T)(this->vn) * ((T)(this->vn + 1) * m4() - (T)3.0 * (T)(this->vn - 1) * m2() * m2());
      return s4 / (T)(this->vn - 1) / (T)(this->vn - 2) / (T)(this->vn - 3);
    }
  };

  template <class T> class statcor // 相関基礎統計量ベースクラス
  {
  protected:                    // 変数をサブクラスに継承するため
    statv<T> xstat;             // xの基礎統計量オブジェクト
    statv<T> ystat;             // yの基礎統計量オブジェクト
    T sigmaxy;                  // Σxy
  public:
    statcor() : xstat(), ystat(), sigmaxy((T)0.0)
    // デフォールトコンストラクタ
    {}
    statcor(T x, T y) : xstat(x), ystat(y), sigmaxy(x * y)
    // 2つのデータによるコンストラクタ
    {}
    statcor(int n0, T sigmax0, T sigmax20, T sigmay0, T sigmay20, T sigmaxy0) : xstat(n0, sigmax0, sigmax20), ystat(n0, sigmay0, sigmay20), sigmaxy(sigmaxy0)
    // 例数等によるコンストラクタ
    {}
    statcor(const statcor& statcor0) : xstat(statcor0.xstat), ystat(statcor0.ystat), sigmaxy(statcor0.sigmaxy)
    // コピーコンストラクタ
    {}
    virtual ~statcor()          // デストラクタ
    {}
    const statcor& operator=(const statcor& statcor0) // コピー代入演算子
    {
      if (&statcor0 == this) return *this; // 自分自身への代入を防ぐ
      xstat = statcor0.xstat;  ystat = statcor0.ystat;
      sigmaxy = statcor0.sigmaxy;
      return *this;             // 自分自身への参照を返す
    }
    const statcor& operator+=(const statcor& statcor0) // 加算代入演算子
    {
      if (statcor0.n() < 1) return *this; // 加算しない
      xstat += statcor0.xstat;  ystat += statcor0.ystat;
      sigmaxy = sigmaxy + statcor0.sigmaxy;
      return *this;             // 自分自身への参照を返す
    }
    void clear()                // データをクリアする
    { xstat.clear();  ystat.clear();  sigmaxy = (T)0.0; }
    int set(T x, T y)           // 2つのデータによる設定
    {
      if (x == (T)BLANKf || y == (T)BLANKf) clear();
      else { xstat.set(x);  ystat.set(y);  sigmaxy = x * y; }
      return n();
    }
    int set(int n0, T mx0, T sdx0, T my0, T sdy0, T rxy0)
    // 例数、平均、標準偏差、相関係数による設定
    {
      if (n0 < 1)
      { clear(); }
      else
      {
        xstat.set(n0, mx0, sdx0);  ystat.set(n0, my0, sdy0);
        T ssxy = rxy0 * sqrt(xstat.ss() * ystat.ss());
        sigmaxy = ssxy + (T)n0 * mx0 * my0;
      }
      return n0;
    }
    virtual int sigma(T x, T y) // 2つのデータを累積する(サブクラスで変更可)
    {
      if (x == (T)BLANKf || y == (T)BLANKf) return 0;
      xstat.sigma(x);  ystat.sigma(y);  sigmaxy += x * y;
      return 1;
    }
    int n() const               // 例数を返す
    { return xstat.n(); }
    T mx() const                // xの平均を返す
    { return xstat.mean(); }
    T sxx() const               // xの平方和を返す
    { return xstat.ss(); }
    T vxx() const               // xの不偏分散を返す
    { return xstat.vr(); }
    T vxxs() const              // xの標本分散を返す
    { return xstat.vrs(); }
    T sdx() const               // xの不偏標準偏差を返す
    { return xstat.sd(); }
    T sdxs() const              // xの標本標準偏差を返す
    { return xstat.sds(); }
    T sex() const               // xの標準誤差を返す
    { return xstat.se(); }
    T my() const                // yの平均を返す
    { return ystat.mean(); }
    T syy() const               // yの平方和を返す
    { return ystat.ss(); }
    T vyy() const               // yの不偏分散を返す
    { return ystat.vr(); }
    T vyys() const              // yの標本分散を返す
    { return ystat.vrs(); }
    T sdy() const               // yの不偏標準偏差を返す
    { return ystat.sd(); }
    T sdys() const              // yの標本標準偏差を返す
    { return ystat.sds(); }
    T sey() const               // yの標準誤差を返す
    { return ystat.se(); }
    T sxy() const               // xとyの積和を返す
    {
      if (n() < 1) return (T)0.0;
      T ssxy = sigmaxy - xstat.sum() * ystat.sum() / (T)(n());
      const T Seps = numeric_limits<T>::epsilon() * fmax((T)FDmin, fabs(sigmaxy));
      return (fabs(ssxy) < Seps) ? (T)0.0 : ssxy;
    }
    T vxy() const               // xとyの不偏共分散を返す
    { return (n() > 1) ? sxy() / (T)(n() - 1) : (T)0.0; }
    T vxys() const              // xとyの標本共分散を返す
    { return (n() > 1) ? sxy() / (T)(n()) : (T)0.0; }
    T rxy() const               // xとyの相関係数を返す
    {
      T sxsy = xstat.ss() * ystat.ss();
      const T Seps = numeric_limits<T>::epsilon() * fmax((T)FDmin, sxsy);
      return (fabs(sxsy) < Seps) ? (T)0.0 : sxy() / sqrt(sxsy);
    }
  };

/* 度数表クラス定義(ソースファイル:ftable.C) */
  class frow                    // 度数表の行クラス
  {
  private:
    string vrcode;              // 行コード
    vecint vfvec;               // 度数ベクトル配列
    bool vcum;                  // 累積行フラッグ
  public:
    frow() : vrcode(), vfvec(1, 0), vcum(false) // デフォールトコンストラクタ
    {}
    frow(size_t nc0, const string& rcode0) : vrcode(rcode0), vfvec(nc0 + 1, 0), vcum(false)
    // コンストラクタ
    {}
    explicit frow(const vecint& intvec) : vrcode(), vfvec(intvec), vcum(false)
    // 度数ベクトル配列型による変換コンストラクタ
    {}
    frow(const frow& frow0) : vrcode(frow0.vrcode), vfvec(frow0.vfvec), vcum(frow0.vcum)
    // コピーコンストラクタ
    {}
    virtual ~frow()             // デストラクタ
    {}
    operator vecint() const     // 度数ベクトル配列型変換演算子
    { return vfvec; }
    const frow& operator=(const frow& frow0); // コピー代入演算子
    const frow& operator=(int f0); // 度数代入演算子(全データ同値代入)
    const frow& operator+=(const frow& frow0); // 加算演算子
    void clear();               // 行をクリアする(初期状態にする)
    size_t erase(size_t cno);   // 行要素を削除する
    size_t erase();    // 全ての行要素を削除する(行コードは削除しない)
    const frow& fclear()        // 全ての度数を0にする
    { return (*this = 0); }
    size_t nc() const           // 列数を返す
    { return vfvec.size() - 1; }
    size_t resize(size_t nc0);  // 列数を変更する
    size_t nc(size_t nc0)       // 列数を変更する
    { return resize(nc0); }
    const string& rcode() const // 行コードを返す
    { return vrcode; }
    const string& rcode(const string& rcode0) // 行コードを設定する
    { return (vrcode = rcode0); }
    int f(size_t cno) const     // 度数を返す
    { return (cno < vfvec.size()) ? vfvec[cno] : -1; }
    int assign(size_t cno, int f0); // 度数を設定する
    int f(size_t cno, int f0)       // 度数を設定する
    { return assign(cno, f0); }
    int ftotal() const          // 度数合計を返す
    { return vfvec[0]; }
    const vecint& fvector() const // 度数ベクトル配列型への参照を返す
    { return vfvec; }
    bool cum() const            // 累積行フラッグを返す
    { return vcum; }
    bool fcum();                // 累積行に変換する
    int get(datac& cdata);      // dansデータから度数を読み込む
  };

  class ftable                  // 度数表クラス
  {
  private:
    list<frow> frowlst;         // 行リスト
    string dcode;               // ダミーコード
    vecstr vccodevec;           // 列コードベクトル配列
    vecint vcwidevec;           // 列コードの桁数ベクトル入れつ
    vector<list<frow>::iterator> ritrvec; // 行リストの反復子ベクトル配列
  public:
    ftable();                   // デフォールトコンストラクタ
    ftable(size_t nr0, size_t nc0); // コンストラクタ
    ftable(const ftable& ftable0);  // コピーコンストラクタ
    virtual ~ftable()               // デストラクタ
    {}
    const ftable& operator=(const ftable& ftable0); // コピー代入演算子
    const ftable& operator=(int f0); // 度数代入演算子(全度数同値代入)
    void clear();               // 度数表をクリアする(初期状態にする)
    size_t rerase(size_t rno);  // 行を削除する
    size_t rerase();            // 全ての行を削除する
    size_t cerase(size_t cno);  // 列を削除する
    size_t cerase();            // 全ての列を削除する
    const ftable& fclear()      // 全ての度数を0にする
    { return (*this = 0); }
    int resize(size_t nr0, size_t nc0); // 行数・列数を変更する
    size_t nr() const                   // 行数を返す
    { return ritrvec.size() - 1; }
    size_t nc() const           // 列数を返す
    { return vccodevec.size() - 1; }
    int nr(size_t nr0)          // 行数を変更する
    { return resize(nr0, nc()); }
    int nc(size_t nc0)          // 列数を変更する
    { return resize(nr(), nc0); }
    const string& hed() const   // ヘッダを返す
    { return vccodevec[0]; }
    const string& hed(const string& hed0) // ヘッダを設定する
    { return (vccodevec[0] = hed0); }
    const string& ccode(size_t cno) const // 列コードを返す
    { return (cno < vccodevec.size()) ? vccodevec[cno] : dcode; }
    const string& ccode(size_t cno, const string& ccode0) // 列コードを設定する
    { return (cno < vccodevec.size()) ? (vccodevec[cno] = ccode0) : dcode; }
    const string& rcode(size_t rno) const // 行コードを返す
    { return (rno < ritrvec.size()) ? ritrvec[rno]->rcode() : dcode; }
    const string& rcode(size_t rno, const string& rcode0) // 行コードを設定する
    { return (rno < ritrvec.size()) ? ritrvec[rno]->rcode(rcode0) : dcode; }
    int hwide() const           // ヘッダの出力桁数を返す
    { return vcwidevec[0]; }
    int hwide(int hwide0)       // ヘッダの出力桁数を設定する
    { return (vcwidevec[0] = hwide0); }
    int cwide(size_t cno) const // 列の出力桁数を返す
    { return (cno < vcwidevec.size()) ? vcwidevec[cno] : -1; }
    int cwide(size_t cno, int cwide0) // 列の出力桁数を設定する
    { return (cno < vcwidevec.size()) ? (vcwidevec[cno] = cwide0) : -1; }
    int twide() const           // 行合計の出力桁数を返す
    { return vcwidevec[nc() + 1]; }
    int twide(int twide0)       // 行合計の出力桁数を設定する
    { return (vcwidevec[nc() + 1] = twide0); }
    int f(size_t rno, size_t cno) const // 度数を返す
    { return (rno < ritrvec.size()) ? ritrvec[rno]->f(cno) : 0; }
    int assign(size_t rno, size_t cno, int f0); // 度数を設定する
    int f(size_t rno, size_t cno, int f0)       // 度数を設定する
    { return assign(rno, cno, f0); }
    int frtotal(size_t rno) const // 行合計を返す
    { return (rno < ritrvec.size()) ? ritrvec[rno]->ftotal() : 0; }
    int fctotal(size_t cno) const // 列合計を返す
    { return ritrvec[0]->f(cno); }
    int ftotal() const          // 全合計を返す
    { return ritrvec[0]->ftotal(); }
    int get(datac& cdata);      // 度数表を読み込む(整理はしない)
    size_t set();               // 度数表を整理し行数を返す
    int trm();                  // 合計が0の行・列を削除する
    bool fcum();                // 累積度数表に変換する
    bool cum() const            // 累積度数表フラッグを返す
    { return ritrvec[0]->cum(); }
  };

  using listftab = list<ftable>;  // 度数表リストの別名定義

  extern int fgetftable(string& fname, listidd& idlst, listftab& ftablst);
  // データ識別情報と度数表を読み込みリストに保存して返す
  extern int putftable(ostream& fout, const ftable& ftab, tabtype tabt = tabfprr);
  // 度数表を度数表形式で出力する

/* 生命表クラス定義(ソースファイル:life.C) */
  class life                    // 生命表クラス
  {
  protected:                    // 変数をサブクラスに継承するため
    int vcno;                   // ケース番号
    string vcode;               // 群コード
    int vcidx;                  // 群コードのインデックス
    double vt;                  // 観察期間
    octype voc;                 // 転帰コード
    double vs;                  // 累積生存率
  public:                       // 
    life() : vcno(0), vcode(), vcidx(-1), vt(0.0), voc(ocnon), vs(1.0)
    // デフォールトコンストラクタ
    {}
    life(int cno0, const string& code0, double t0, octype oc0) : vcno(cno0), vcode(code0), vcidx(-1), vt(t0), voc(oc0), vs(1.0)
    // コンストラクタ
    {}
    life(const life& life0) : vcno(life0.vcno), vcode(life0.vcode), vcidx(life0.vcidx), vt(life0.vt), voc(life0.voc), vs(life0.vs)
    // コピーコンストラクタ
    {}
    virtual ~life()             // デストラクタ
    {}
    virtual const life& operator=(const life& life0); // コピー代入演算子
    virtual bool operator<(const life& life0) const;  // 比較演算子
    int cno() const                           // ケース番号を返す
    { return vcno; }
    const string& code() const  // 群コードを返す
    { return vcode; }
    int cidx() const            // 群コードのインデックスを返す
    { return vcidx; }
    int cidx(int cidx0)         // 群コードのインデックスをセットする
    { return (vcidx = cidx0); }
    double t() const            // 観察期間を返す
    { return vt; }
    octype oc() const           // 転帰コードを返す
    { return voc; }
    double s() const            // 累積生存率を返す
    { return vs; }
    double s(double s0)         // 累積生存率を設定する
    { return (vs = s0); }
  };

  class lifem : public life     // 多変量生命表クラス
  {
  protected:                    // 変数をサブクラスに継承するため
    vecdbl vxvec;               // 共変数ベクトル
  public:
    lifem() : life(), vxvec()   // デフォールトコンストラクタ
    {}
    lifem(int cno0, double t0, octype oc0, const vecdbl& xvec0) : life(cno0, BLANKs, t0, oc0), vxvec(xvec0)
    // コンストラクタ
    {}
    lifem(int cno0, const string& code0, double t0, octype oc0, const vecdbl& xvec0) : life(cno0, code0, t0, oc0), vxvec(xvec0)
    // コンストラクタ
    {}
    lifem(const lifem& life0) : life(life0), vxvec(life0.vxvec)
    // コピーコンストラクタ
    {}
    virtual ~lifem()            // デストラクタ
    {}
    virtual const lifem& operator=(const lifem& life0); // コピー代入演算子
    double x(int idx) const     // 共変数を返す
    { return vxvec[idx]; }
  };

  using listlf = list<life>;             // 生命表リストの短名定義
  using listlf_itr = list<life>::iterator; // 生命表リスト反復子の短名定義
  using listlf_citr = list<life>::const_iterator; // 生命表リスト反復子の短名定義
  using listlfm = list<lifem>;               // 多変量生命表リストの短名定義
  using listlfm_itr = list<lifem>::iterator; // 多変量生命表リスト反復子の短名定義
  using listlfm_citr = list<lifem>::const_iterator; // 多変量生命表リスト反復子の短名定義

/* dans用共通関数群(ソースファイル:dfun.C) */
  extern int fgetdata(string& fname, idd& id, ldata& datalst);
  // ケースデータを読み込みデータリストに保存する
  extern double strtodf(const string& str); // 文字列を倍精度実数に変換する
  extern const string& strtodstr(string& str, bool squot = false);
  // 文字列をdans用特殊表現に変換する
  extern int search_data(const set<string>& sset, const string& sdata);
  // 文字セット中の文字データをサーチしてインデックスを返す
  extern int search_data(const set<double>& fset, double fdata);
  // 実数セット中の実数データをサーチしてインデックスを返す
  extern int search_data(const set<pair<string, double>>& targset, const string& sdata);
  // 文字と実数セット中の文字データをサーチしてインデックスを返す
  extern int search_data(const set<pair<string, double>>& targset, double fdata);
  // 文字と実数セット中の実数データをサーチしてインデックスを返す
  extern testtype ctest(double& csqr, int& df, const ftable& ftab, bool ys = true);
  // χ2乗検定
  extern double csqr22(int a, int b, int c, int d); // 2x2のχ2乗値を返す
  extern testtype utest(double& u, double& vu, double& z, const ftable& ftab, testtype ut = testaut);
  // Mann-WhitneyのU検定(Wilcoxonの順位和検定)
  extern int htest(double& csqr, double& kc, vecdbl& tvec, const ftable& ftab);
  // Kruskal-WallisのH検定
  extern double spear(const ftable& ftab);  // spearmanの順位相関係数計算
  extern double kappa(double& vk, double& vk0, const ftable& ftab, bool weight);
  // 一致係数κの計算
  extern double jttest(const ftable& ftab, double& jt, double& tmax, double& vt);
  // Jonckheere-Terpstra検定計算
  extern double ftest(const vecdbl& pvec, const ftable& ftab, bool ys = false);
  // 出現率一様性の検定
  extern double cctest(double& csqr, double& v, const ftable& ftab);
  // 累積χ2乗検定
  extern ldbl cctop(ldbl csqr, ldbl v); // χ2乗値・小数自由度→p値変換
  extern int rrtest(double& rr, double& vrr, double& ca, double& ch, const listftab& ftablst);
  // リスク比検定
  extern int lgtest(double& odr, double& vodr, double& ca, double& ch, const listftab& ftablst);
  // ロジットによるオッズ比の検定
  extern int mbtest(double& orm, double& cmm, double& cmr, const listftab& ftablst, testtype ct);
  // Mantel-Haenzelの検定+Breslow-Dayの検定
  extern int mtest(double& rdm, double& vrdm, double& rrm, double& rra, double& rrb, double& rrc, double& orm, double& ora, double& orb, double& orc, double& cmm, double& cmr, const listftab& ftablst, testtype ct);
  // Mantel-Haenzelの検定
  extern int emtest(double& cmm, double& cmr, double& rd, double& vrd, const listftab& ftablst);
  // 拡張Mantel検定
  extern int gemtest(double& cmm, double& cmr, const listftab& ftablst);
  // 一般化拡張Mantel検定
  extern double wtest(const ftable& ftab, double& tp, double& vt, testtype wt);
  // Wilcoxonの符号付き順位検定
  extern ldbl calmcpr(mctype mc, double& f, int g, int df1, int df2, int dc, double r = 0.5);
  // 多重比較の有意確率を計算する(df2==0:ノンパラ)
  extern double calmcci(mctype mc, double cc, int g, int df1, int df2, int dc, double v, double r = 0.5);
  // 多重比較の信頼区間を計算する(df2==0:ノンパラ)

  extern double calbexci(double& plow, double& pup, int n, int nt, double cc);
  // F分布を利用した出現率の信頼区間を返す
  extern double calbnmci(double& plow, double& pup, double p0, int nt, double cc, bool cor = true);
  // 正規近似を利用した出現率の信頼区間を返す
  extern double caldpci(double& d, mctype mc, double cc, int g, int df, int dc, int n1, int n2, int m1, int m2, bool cor = true, double r = 0.5);
  // 出現率の差の信頼区間を計算する(多重比較含む)
  extern int calfci(double& fl, double& fu, double cc, int df1, int df2, double f);
  // 分散比Fの信頼区間を返す
  extern int calcrci(double& crl, double& cru, double cc, int df1, int df2, double f);
  // 相関比の信頼区間を返す
  extern int calcsqrci(double& csqrl, double& csqru, double cc, int df, double csqr);
  // χ2乗値の信頼区間を返す
  extern double calrci(double& zr, double cc, int n, double r);
  // 相関係数の信頼区間を返す
  extern void calelp(double& a, double& b, double& h, double cc, int df1, int df2, double vxx, double vyy, double vxy);
  // 信頼楕円のパラメータを返す
  extern int setpcvec(vecint& pcvec, const lstr& klst, int pmax);
  // 周期成分番号を周期成分番号ベクトル配列にセットする

/* 統計関数群(ソースファイル:sfun.C) */
  extern ldbl ctop(ldbl csqr, int df); // χ2乗値→上側p値変換
  extern ldbl ztop(ldbl z);            // 正規分布z値→両側p値変換
  extern ldbl nedp(ldbl ned);          // 正規偏位→下側p値変換
  extern ldbl ftop(ldbl f, int df1, int df2); // F値→上側p値変換
  extern ldbl ttop(ldbl t, int df = 0); // t値→両側p値変換
  extern ldbl qtop(ldbl q, int g, int df = 0);
// スチューデント化された範囲のq値→上側p値変換
  extern ldbl dtop(ldbl d, int g, int df = 0, ldbl r = 0.5L);
// ダネットのd値→両側p値変換(多変量t分布利用)
  extern ldbl ncctolp(ldbl csqr, int df, ldbl lamda = 0.0L);
// 非心χ2乗値→下側p値変換
  extern ldbl ncttolp(ldbl t, int df, ldbl lamda = 0.0L);
// 非心t値→下側p値変換
  extern ldbl ncftolp(ldbl f, int df1, int df2, ldbl lamda = 0.0L);
// 非心F値→下側p値変換
  extern ldbl guessz(ldbl pr);  // 正規分布%点,Hesstingの最良近似式
  extern ldbl ptoc(ldbl pr, int df); // 上側p値→χ2乗値逆変換
  extern ldbl ptoz(ldbl pr);         // 両側p値→正規分布z値逆変換
  extern ldbl ptoned(ldbl pr);       // 下側p値→正規偏位逆変換
  extern ldbl ptof(ldbl pr, int df1, int df2); // 上側p値→F値逆変換
  extern ldbl ptot(ldbl pr, int df = 0);       // 両側p値→t値逆変換
  extern ldbl ptoq(ldbl pr, int g, int df = 0);
// 上側p値→スチューデント化された範囲のq値逆変換
  extern ldbl ptod(ldbl pr, int g, int df = 0, ldbl r = 0.5L);
// 両側p値→ダネットのq値逆変換
  extern ldbl lptoncc(ldbl pr, int df, ldbl lamda = 0.0L);
// 下側p値→非心χ2乗値逆変換
  extern ldbl lptonct(ldbl pr, int df, ldbl lamda = 0.0L);
// 下側p値→非心t値逆変換
  extern ldbl lptoncf(ldbl pr, int df1, int df2, ldbl lamda = 0.0L);
// 下側p値→非心F値逆変換
  extern ldbl snden(ldbl z);     // 標準正規分布密度関数
  extern ldbl sndis(ldbl z);     // 標準正規分布関数
  extern ldbl cden(ldbl c, int df); // χ2乗分布密度関数
  extern ldbl cdis(ldbl c, int df); // χ2乗分布関数
  extern ldbl tden(ldbl t, int df); // t分布密度関数
  extern ldbl tdis(ldbl t, int df); // t分布関数
  extern ldbl fden(ldbl f, int df1, int df2); // F分布密度関数
  extern ldbl fdis(ldbl f, int df1, int df2); // F分布関数
  extern ldbl bnmden(int a, int n, ldbl p0 = 0.5L); // 二項分布密度関数
  extern ldbl bnmdis(int a, int n, ldbl p0 = 0.5L); // 二項分布関数
  extern ldbl binom(int a, int b, ldbl p0 = 0.5L); // 二項検定
  extern ldbl fisden(int a, int b, int c, int d); // 超幾何分布密度関数
  extern ldbl fisdis(int a, int b, int c, int d); // 超幾何分布関数
  extern ldbl fisher(int a, int b, int c, int d); // Fisherの正確検定
  extern ldbl logden(ldbl t, ldbl m, ldbl c, ldbl y0);
// ロジスティック分布密度関数
  extern ldbl logdis(ldbl t, ldbl m, ldbl c, ldbl y0);
// ロジスティック分布関数
  extern ldbl mnden(const vecldbl& zvec, int p, const mat<ldbl>& rrmat, ldbl rdet);
// 多変量正規分布密度関数
  extern const string pind(ldbl pr); // 有意確率p値→有意の記号を返す
  extern double probit(double pr);     // 反応率→プロビット変換
  extern double randn(double m = 0.0, double sd = 1.0);
// N(m, sd2乗)の正規乱数を返す

/* 統計量出力関数群(ソースファイル:putstat.C) */
  extern int puttitle(ostream& fout, const string& title, int wline = LENcrt);
// DANSのタイトルとバージョンを出力する
  extern double putatab(ostream& fout, const string& factor, double sa, int dfa, double sr, int dfr, double st, int dft);
// 分散分析表を出力して残差分散を返す
  extern double calvfp(double& v, double& f, double& p, double ss, int df, double sr, int dfr);
// 分散、F値、p値を計算して分散を返す
  extern void putmsd(ostream& fout, int n, double mean, double v);
// 例数、平均値、標準偏差を出力する

/* dans用周期関数クラス定義(テンプレートクラス) */
  template <class T> class pfun // 周期関数クラス
  {
  private:
    vector<pair<T, T>> vabpvec; // 周期関数係数ベクトル配列
    vector<pair<T, T>> vaapvec; // 振幅・位相ベクトル配列(ラジアン単位)
    vecint vpcvec;              // 周期成分番号ベクトル配列
    int vpcmax;                 // 周期成分番号最大値
    T vt0;                      // 基本周期
    T vw0;                      // 角周波数ω0=2π/T
  public:
    pfun() : vabpvec(), vaapvec(), vpcvec(), vpcmax(-1), vt0((T)0.0), vw0((T)0.0)
    // デフォールトコンストラクタ
    {}
    pfun(T t0, const vecint& pcvec0, const vector<T>& abvec0) : vpcvec(pcvec0), vpcmax(-1), vt0(t0), vw0((T)2.0 * (T)M_PI / t0)
    // コンストラクタ
    {
      size_t m = pcvec0.size();
      vabpvec.resize(m + 1);  vaapvec.resize(m + 1);
      vabpvec[0].first = abvec0[0];  vabpvec[0].second = (T)0.0;
      vaapvec[0].first = abvec0[0];  vaapvec[0].second = (T)0.0;
      for (size_t k = 1; k <= m; ++k)
      {
        size_t i = k * 2;
        T a = (i - 1 < abvec0.size()) ? abvec0[i - 1] : (T)0.0;
        T b = (i < abvec0.size()) ? abvec0[i] : (T)0.0;
        T aabs = fabs(a), babs = fabs(b);
        const T ABeps = numeric_limits<T>::epsilon() * fmax((T)FDmin, fmax(aabs, babs));
        if (aabs < ABeps) a = (T)0.0;
        if (babs < ABeps) b = (T)0.0;
        vabpvec[k].first = a;  vabpvec[k].second = b;
        vaapvec[k].first = hypot(a, b);
        T h = atan2(b, a);
        const T Heps = numeric_limits<T>::epsilon() * fmax((T)FDmin, fabs(h));
        if (fabs(h) < Heps) h = (T)0.0;
        else if (h < 0.0L) h += (T)2.0 * (T)M_PI;
        vaapvec[k].second = h;
      }
    }
    pfun(T t0, const vecint& pcvec0, const vector<pair<T, T>>& aapvec0) : vpcvec(pcvec0), vaapvec(aapvec0), vpcmax(-1), vt0(t0), vw0((T)2.0 * (T)M_PI / t0)
    // コンストラクタ
    {
      size_t m = pcvec0.size();
      vabpvec.resize(m + 1);
      size_t ms = vaapvec.size();
      if (ms < m + 1)
      {
        vaapvec.resize(m + 1);
        for (size_t k = ms; k <= m; ++k) vaapvec[k].first = vaapvec[k].second = (T)0.0;
      }
      vabpvec[0].first = vaapvec[0].first;
      vabpvec[0].second = vaapvec[0].second;
      for (size_t k = 1; k <= m; ++k)
      {
        T am = vaapvec[k].first, ac = vaapvec[k].second;
        T amabs = fabs(am), acabs = fabs(ac);
        const T Aeps = numeric_limits<T>::epsilon() * fmax((T)FDmin, fmax(amabs, acabs));
        if (amabs < Aeps) am = (T)0.0;
        if (acabs < Aeps) ac = (T)0.0;
        T tanth = tan(ac);
        T a = am / sqrt((T)1.0 + tanth * tanth);
        T b = a * tanth;
        vabpvec[k].first = a;  vabpvec[k].second = b;
      }
    }
    pfun(const pfun<T>& pfun0) : vabpvec(pfun0.vabpvec), vaapvec(pfun0.vaapvec), vpcvec(pfun0.vpcvec), vpcmax(pfun0.vpcmax), vt0(pfun0.vt0), vw0(pfun0.vw0)
    // コピーコンストラクタ
    {}
    virtual ~pfun()             // デストラクタ
    {}
    const pfun<T>& operator=(const pfun<T>& pfun0) // コピー代入演算子
    {
      if (&pfun0 == this) return *this; // 自分自身への代入を防ぐ
      vpcvec = pfun0.vpcvec;
      vabpvec = pfun0.vabpvec;  vaapvec = pfun0.vaapvec;
      vpcmax = pfun0.vpcmax;  vt0 = pfun0.vt0;  vw0 = pfun0.vw0;
      return *this;
    }
    void clear()                // 要素を削除する
    {
      vpcvec.clear();  vabpvec.clear();  vaapvec.clear();
      vt0 = (T)0.0;  vw0 = (T)0.0;  vpcmax = -1;
    }
    T a(size_t k) const         // cos項の係数akを返す
    { return (k <= vpcvec.size()) ? vabpvec[k].first : (T)0.0; }
    T b(size_t k) const         // sin項の係数bkを返す
    { return (k <= vpcvec.size()) ? vabpvec[k].second : (T)0.0; }
    T amp(size_t k) const       // 振幅Akを返す
    { return (k <= vpcvec.size()) ? vaapvec[k].first : (T)0.0; }
    T acr(size_t k) const       // 位相θk(ラジアン単位)を返す
    { return (k <= vpcvec.size()) ? vaapvec[k].second : (T)0.0; }
    T dfun(T t) const           // 周期関数の微分係数を返す
    {
      T dy = (T)0.0;
      for (size_t k = 0; k < vpcvec.size(); ++k)
      {
        T wk = vw0 * (T)(vpcvec[k]);
        dy += wk * (vabpvec[k + 1].second * cos(wk * t) - vabpvec[k + 1].first * sin(wk * t));
      }
      return dy;
    }
    T d2fun(T t) const          // 周期関数の2階微分係数を返す
    {
      T d2y = (T)0.0;
      for (size_t k = 0; k < vpcvec.size(); ++k)
      {
        T wk = vw0 * (T)(vpcvec[k]);
        d2y -= wk * wk * (vabpvec[k + 1].first * cos(wk * t) + vabpvec[k + 1].second * sin(wk * t));
      }
      return d2y;
    }
    T fun(T t) const            // 周期関数値を返す
    {
      T y = vabpvec[0].first;
      for (size_t k = 0; k < vpcvec.size(); ++k)
      {
        T tk = vw0 * (T)(vpcvec[k]) * t;
        y += vabpvec[k + 1].first * cos(tk) + vabpvec[k + 1].second * sin(tk);
      }
      return y;
    }
    T integ(T ts, T te) const   // 周期関数の定積分を返す
    {
      T y = vabpvec[0].first * (te - ts);
      for (size_t k = 0; k < vpcvec.size(); ++k)
      {
        T wk = vw0 * (T)(vpcvec[k]);
        y += vabpvec[k + 1].first / wk * (sin(wk * te) - sin(wk * ts)) - vabpvec[k + 1].second / wk * (cos(wk * te) - cos(wk * ts));
      }
      return y;
    }
    size_t m() const            // 周期成分数を返す
    { return vpcvec.size(); }
    int pcmax()                 // 周期成分番号最大値を返す
    {
      if (vpcmax >= 0) return vpcmax; // サーチ済み
      for (size_t k = 0; k < vpcvec.size(); ++k)
      {
        if (vpcvec[k] > vpcmax) vpcmax = vpcvec[k];
      }
      return vpcmax;
    }
    int pcn(size_t k) const     // 周期成分番号を返す
    { return (0 < k && k <= vpcvec.size()) ? vpcvec[k - 1] : 0; }
    T t(size_t k) const         // 周期Tkを返す
    { return (0 < k && k <= vpcvec.size()) ? vt0 / vpcvec[k - 1] : vt0; }
    T w(size_t k) const         // 角周波数ωk(ラジアン単位)を返す
    { return (0 < k && k <= vpcvec.size()) ? vw0 * vpcvec[k - 1] : vw0; }
  };

  template<class T> int settvec(vector<T>& tvec, const lstr& tlst, int pmax)
  // 項目時間を項目時間ベクトル配列にセットする
  {
    if (tlst.empty()) return 0;
    if (tlst.size() > (size_t)pmax) return -1; // 項目時間指定が多すぎる
    size_t j = 0;
    for (auto tstr : tlst)
    {
      T fdata = (T)(strtodf(tstr));
      if (fdata == (T)BLANKf) return -2; // 時間指定が間違い
      if (j > 0 && fdata <= tvec[j - 1]) return -3; // 前の時間より大きくない
      tvec[j] = fdata;
      ++j;
    }
    return (int)(tlst.size());  // 項目時間指定数を返す
  }

/* 多変量の統計量出力用関数群(テンプレート関数) */
  template <class T> void putmsdvec(ostream& fout, int n, const vector<T>& mevec, const vector<T>& vrvec, size_t xidx = 0, size_t yidx = 0)
  // 多変数の基礎統計量を出力する
  {
    if (mevec.empty() || vrvec.empty()) return; // 出力なし
    size_t pidx = mevec.size() - 1;
    if (xidx > 0) pidx = xidx + yidx;
    else xidx = pidx;
    if (n < 1 || pidx < 1) return; // 出力なし
    fout << "\n<BASIC STATISTICS OF VARIABLES>\n";
    constexpr int LENline = 70;
    putline(fout, LENline);
    if (yidx == 1)
    {
      fout << "Y   :";
      putmsd(fout, n, (T)(mevec[pidx]), (T)(vrvec[pidx]));
      pidx = xidx;
    }
    for (size_t j = 1; j <= pidx; ++j)
    {
      if (j <= xidx)
      { fout << "X" << setw(3) << j; }
      else if (yidx > 1)
      { fout << "Y" << setw(3) << j - xidx; }
      fout << ":";  putmsd(fout, n, (T)(mevec[j]), (T)(vrvec[j]));
    }
    putline(fout, LENline);
  }

  template <class T> void putmat(ostream& fout, const string& head, mattype mtype, int n, const mat<T>& smat, size_t xidx = 0, size_t yidx = 0)
  // 各種行列を出力する
  {
    if (smat.r() < 1) return;   // 出力なし
    constexpr size_t Cmax = 9;  // 行列表横の最大個数
    int df = n - 1;
    size_t pidx = smat.r() - 1;
    if (xidx > 0) pidx = xidx + yidx;
    else xidx = pidx;
    if (df < 0 || pidx < 1) return; // 出力なし
    int f = 1, fmax = 0;
    for (size_t i = 1; i <= pidx; ++i)
    {
      if (smat(i, i) > (T)0.0) f = (int)(log10(smat(i, i))) + 1;
      else if (smat(i, i) < (T)0.0) f = (int)(log10(-smat(i, i))) + 2;
      if (fmax < f) fmax = f;
    }
    f = (fmax < 6) ? 6 - fmax : 0; // 小数部桁数を最大6桁にする
    size_t low = 1;
    T val = (T)0.0;
    fout << "\n" << head;
    while (low <= pidx)
    {
      size_t up = (low + Cmax - 1 < pidx) ? low + Cmax - 1 : pidx;
      fout << "\n    ";
      for (size_t i = low; i <= up; ++i)
      {
        if (i <= xidx) fout << "    X" << setw(3) << i;
        else fout << "    Y" << setw(3) << i - xidx;
      }
      fout << "\n";
      size_t llen = 4 + 8 * (up - low + 1);
      putline(fout, llen);
      int prec0 = fout.precision(); // 現在の小数点以下の精度を保存する
      for (size_t i = 1; i <= pidx; ++i)
      {
        if (i <= xidx) fout << "X" << setw(3) << i;
        else fout << "Y" << setw(3) << i - xidx;
        fout.setf(ios::fixed);
        for (size_t j = low; j <= up; ++j)
        {
          switch (mtype)
          {
          case mats : fout << setw(8) << setprecision(f) << smat(i, j);
            break;
          case matv : val = (df > 0) ? smat(i, j) / (T)df : (T)0.0; // 共分散行列
            fout << setw(8) << setprecision(f) << val;
            break;
          case matr :             // 相関行列
          default : val = smat(i, i) * smat(j, j);
            val = (val > (T)0.0) ? smat(i, j) / sqrt(val) : (T)0.0;
            if (i == j) fout << "    1   ";
            else fout << setw(8) << setprecision(3) << val;
          }
        }
        fout.unsetf(ios::fixed);
        fout << "\n";
      }
      fout << setprecision(prec0); // 小数点以下の精度を元に戻す
      putline(fout, llen);
      low += Cmax;
    }
  }

  template <class T> void putcoef(ostream& fout, const string& mtitle, const string& rhed, const string& ched, const mat<T>& amat)
  // 多変量解析各種スコア計算用係数を出力する
  {
    constexpr size_t Cmax = 6;  // 行列表横の最大数
    size_t pidx = amat.r() - 1, qidx = amat.c() - 1, low = 1, up = 0;
    fout << "\n" << mtitle;
    while (low <= qidx)
    {
      up = (low + Cmax - 1 < qidx) ? low + Cmax - 1 : qidx;
      fout << "\nVAR ";
      for (size_t j = low; j <= up; ++j)
      {
        int npad = rhed.size() - strclen(rhed, CHARwide);
        fout << setw(9 + npad) << rhed;
        fout << setw(3) << j;
      }
      fout << "\n";
      size_t llen = 4 + 12 * (up - low + 1);
      putline(fout, llen);
      int prec0 = fout.precision(); // 現在の小数点以下の精度を保存する
      for (size_t i = 0; i <= pidx; ++i)
      {
        if (i == 0)
        { fout << "CNST"; }
        else
        {
          fout << setw(1) << ched;
          fout << setw(3) << i;
        }
        for (size_t j = low; j <= up; ++j)
        {
          if (fabs(amat(i, j)) >= (T)100.0)
          { fout << setw(12) << amat(i, j); }
          else
          {
            fout.setf(ios::fixed);
            fout << setw(12) << setprecision(7) << amat(i, j) << setprecision(prec0);
            fout.unsetf(ios::fixed);
          }
        }
        fout << "\n";
      }
      putline(fout, llen);
      low = up + 1;
    }
  }

  template <class T> void putfl(ostream& fout, const string& mtitle, const string& rhed, const string& ched, mat<T>& lmat)
  // 因子負荷量を出力する
  {
    constexpr size_t Cmax = 6;  // 行列表横の最大数
    size_t pidx = lmat.r() - 1, qidx = lmat.c() - 1, low = 1, up = 0;
    bool eop = false;           // 表出力終了フラッグ
    for (size_t i = 0; i <= pidx; ++i) lmat.assign(i, 0, (T)0.0);
    fout << "\n" << mtitle;
    while (!eop)
    {
      size_t nc = low + Cmax - 1;
      if (low > qidx) eop = true;
      else if (nc > qidx) { up = qidx;  eop = true; }
      else if (nc == qidx) up = qidx;
      else up = nc;
      fout << "\nVAR ";
      for (size_t j = low; j <= up; ++j)
      {
        int npad = rhed.size() - strclen(rhed, CHARwide);
        fout << setw(9 + npad) << rhed;
        fout << setw(3) << j;
      }
      if (eop) fout << "         r^2";
      fout << "\n";
      size_t llen = 4 + 12 * (up - low + 1 + (eop ? 1 : 0));
      putline(fout, llen);
      for (size_t i = 1; i <= pidx; ++i)
      {
        fout << setw(1) << ched;
        fout << setw(3) << i;
        int prec0 = fout.precision();
        fout.precision(6);
        fout.setf(ios::fixed);
        for (size_t j = low; j <= up; ++j)
        {
          fout << setw(12) << lmat(i, j);
          lmat.assign(i, 0, lmat(i, 0) + lmat(i, j) * lmat(i, j));
        }
        if (eop) { fout << setw(12) << lmat(i, 0); }
        fout << "\n";
        fout.unsetf(ios::fixed);
        fout.precision(prec0);
      }
      putline(fout, llen);
      low = up + 1;
    }
  }

  template <class T> void smat2rmat(mat<T>& smat)
  // 積和行列を相関行列に変換する
  {
    for (size_t i = 0; i < smat.r(); ++i)
    {
      T sii = smat(i, i);
      for (size_t j = i + 1; j < smat.c(); ++j)
      {
        T ss = sii * smat(j, j);
        smat.assign(i, j, (ss > (T)0.0) ? smat(i, j) / sqrt(ss) : (T)0.0);
        smat.assign(j, i, smat(i, j));
      }
      smat.assign(i, i, (T)1.0);
    }
  }

} // 名前空間dans終わり

#endif
