// datace.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2023-10-27 17:27:47 tac>
//
// データクラスのメンバー関数
//
// ライブラリ:libdanse.so.7.4
// コンパイル法:g++ -O -fPIC -c datace.C

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const datac& datac::operator=(const datac& data0) // コピー代入演算子
{
  if (&data0 == this) return *this; // 自分自身への代入を防ぐ
  blank = data0.blank;
  vdstr = data0.vdstr;  dstridx = data0.dstridx;
  vfdata = data0.vfdata;  fdataidx = data0.fdataidx;
  vsdatavec = data0.vsdatavec;
  return *this;
}

const datac& datac::operator=(const string& sdata0)
// 文字データ代入演算子(全データ同値代入)
{
  vsdatavec.assign(vsdatavec.size(), sdata0);
  dstridx = fdataidx = UINT_MAX;  vdstr = BLANKs;  vfdata = BLANKf;
  return *this;
}

const datac& datac::operator=(double fdata0)
// 実数データ代入演算子(全データ同値代入)
{
  ostringstream ostr;
  if (fdata0 == BLANKf) ostr << BLANKs; // 文字型ブランクデータにする
  else ostr << fdata0;
  vsdatavec.assign(vsdatavec.size(), ostr.str());
  dstridx = UINT_MAX;  vdstr = BLANKs;
  fdataidx = 0;  vfdata = fdata0; // 実数データを最初のデータとしてセットする
  return *this;
}

void datac::clear()             // 全データを削除する
{
  blank = BLANKs;  vdstr = BLANKs;  dstridx = UINT_MAX;
  vfdata = BLANKf;  fdataidx = UINT_MAX;  vsdatavec.clear();
}

int datac::erase(size_t idxbeg, size_t idxend)
// 指定された範囲のデータを削除する
{
  if (idxend == UINT_MAX) idxend = idxbeg; // 1つだけ削除
  if (vsdatavec.size() <= idxend) return -1; // 削除せず
  vecstr_itr pos0 = vsdatavec.begin();
  vecstr_itr pos = vsdatavec.erase(pos0 + idxbeg, pos0 + idxend);
  if (idxbeg <= dstridx) dstridx = UINT_MAX;
  if (idxbeg <= fdataidx) fdataidx = UINT_MAX;
  return idxbeg;                // 次の要素のインデックスを返す
}

size_t datac::resize(size_t n0) // データ数を変更する
{
  if (n0 == vsdatavec.size()) return n0; // 変更不要
  vsdatavec.resize(n0, BLANKs);
  if (dstridx >= n0) dstridx = UINT_MAX; // 特殊文字データのインデックスをクリアする
  if (fdataidx >= n0) fdataidx = UINT_MAX; // 実数データのインデックスをクリアする
  return n0;                    // 変更後のデータ数を返す
}

const string& datac::assign(size_t idx, const string& sdata0)
// 文字データを設定する
{
  if (vsdatavec.size() <= idx) return blank; // 変更できず
  vsdatavec[idx] = sdata0;
  if (idx == dstridx) dstridx = UINT_MAX; // 特殊文字データのインデックスをクリアする
  if (idx == fdataidx) fdataidx = UINT_MAX; // 実数データのインデックスをクリアする
  return vsdatavec[idx];        // 変更後のデータを返す
}

double datac::assign(size_t idx, double fdata0)
// 実数データを設定する
{
  constexpr int PRECdef = 16;                 // 省略時浮動小数点精度
  if (vsdatavec.size() <= idx) return BLANKf; // 変更できず
  ostringstream ostr;
  ostr.precision(PRECdef);
  if (fdata0 == BLANKf) ostr << BLANKs; // 文字型ブランクデータにする
  else ostr << fdata0;
  vsdatavec[idx] = ostr.str();
  if (idx == dstridx) dstridx = UINT_MAX; // 特殊文字データのインデックスをクリアする
  vfdata = fdata0;  fdataidx = idx; // 実数データとインデックスをセットする
  return vfdata;                // 変更後のデータを返す
}

const string& datac::dstr(size_t idx) // 特殊文字列を用いた表示用データを返す
{
  if (vsdatavec.size() <= idx) return blank;
  if (idx == dstridx) return vdstr; // 現在のインデックスを指定された時
  vdstr = vsdatavec[idx];  dstridx = idx;
  strtodstr(vdstr);             // dans用特殊表現に変換する
  return vdstr;
}

double datac::fdata(size_t idx) // 実数データを返す
{
  if (vsdatavec.size() <= idx) return BLANKf;
  if (idx == fdataidx) return vfdata; // 現在のインデックスを指定された時
  vfdata = strtodf(vsdatavec[idx]);  fdataidx = idx;
  return vfdata;
}

int datac::insert(size_t idx, const string& sdata0) // 文字データを挿入する
{
  if (idx > vsdatavec.size()) return -1; // 挿入できず
  if (idx == vsdatavec.size())  // データを末尾に追加する
  { vsdatavec.push_back(sdata0); }
  else
  {
    vecstr_itr pos0 = vsdatavec.begin();
    vecstr_itr pos = vsdatavec.insert(pos0 + idx, sdata0);
  }
  if (idx <= dstridx) dstridx = UINT_MAX; // 特殊文字データのインデックスをクリアする
  if (idx <= fdataidx) fdataidx = UINT_MAX; // 実数データのインデックスをクリアする
  return idx;                   // 挿入後のインデックスを返す
}

int datac::put(ostream& fout)   // ケースデータをファイルに出力する
{
  if (vsdatavec.size() > 0)
  {
    fout << dstr(0);
    for (size_t i = 1; i < vsdatavec.size(); ++i) fout << CHRdelm << dstr(i);
  }
  fout << endl;
  return vsdatavec.size();      // 出力したデータ数を返す
}

int datac::getdata(const lstr& strlst)
// 文字列リストからデータを取り出し、取り出したデータ数を返す
{
  vdstr = BLANKs;  dstridx = UINT_MAX;  vfdata = BLANKf;  fdataidx = UINT_MAX;
  if (vsdatavec.size() < 1) vsdatavec.resize(strlst.size());
  lstr_citr spos = strlst.begin();
  size_t i = 0;
  while (i < vsdatavec.size() && spos != strlst.end())
  {
    vsdatavec[i] = *spos;
    strtrm(vsdatavec[i]);         // 前後の空白を削除する
    sstrtostr(vsdatavec[i]);      // 特殊表現文字を変換し""を削除する
    ++spos;  ++i;
  }
  while (i < vsdatavec.size()) vsdatavec[i++] = BLANKs;
  return strlst.size();         // 取り出したデータ数を返す
}

int datac::getdata(const string& str, char delmc)
// 文字列からデータを取り出し、取り出したデータ数を返す
{
  lstr strlst;
  strsplit(strlst, str, delmc);
  return getdata(strlst);       // 取り出したデータ数を返す
}
