// dfune.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-08 09:17:41 tac>
//
// 共通関数群
//
// ライブラリ:libdanse.so.7.4
// コンパイル法:g++ -O -fPIC -c dfune.C

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

int dans::fgetdata(string& fname, idd& id, ldata& datalst)
// ケースデータを読み込みデータリストに保存する
{
  istream* finp = nullptr;
  if (fname.empty())
  { finp = &cin;  fname = "stdin"; }
  else
  {
    finp = new ifstream(fname.c_str());
    if (!(*finp)) return disperr(fname, "No such file !");
  }
  string sline;                 // データ入力用
  datac cdata(id.ni());         // データオブジェクト
  while (getsline(sline, *finp) != EOF)
  {
    strchomp(sline);
    if (sline[0] == CHRtags)    // データ識別情報
    {
      if (id.append(sline) < 0) return disperr(fname, string("Wrong data identification information->") + sline);
    }
    else                        // ケースデータ読み込み
    {
      if (id.nc() < 1)          // 最初の読み込み
      {
        if (id.ni() < 1)        // データ識別情報がない場合
        {
          int ni = cdata.getdata(sline);
          if (ni < 1) return disperr(fname, string("No data on the first line !->") + sline);
          else id.set(fname, ni);
        }
        else
        {
          cdata.resize(id.ni());
          cdata.getdata(sline);
        }
      }
      else
      { cdata.getdata(sline); }
      datalst.push_back(cdata);  id.nc(id.nc() + 1);
    }
  }
  if (finp != &cin) delete finp;
  return (int)(datalst.size()); // 読み込んだケース数を返す
}

double dans::strtodf(const string& str) // 文字列を倍精度実数に変換する
{
  if (str.empty() || str == BLANKs) return BLANKf;
  char* sptr = nullptr;
  double d = strtod(str.c_str(), &sptr);
  if (sptr != 0 && *sptr != '\0') return BLANKf;
  return d;
}

const string& dans::strtodstr(string& str, bool squot)
// 文字列をDANS用特殊表現に変換する
{
  if (str.size() < 1) return str;
  bool quot = false;
  if (str.find(CHRtags) != string::npos)
  { quot = true; }
  else if (str.find(CHRtage) != string::npos)
  { quot = true; }
  else if (str.find(CHRdelm) != string::npos)
  { quot = true; }
  else
  {
    string tstr(str);
    if (size_t(strtrm(tstr)) < str.size()) quot = true; // 前後に空白がある
  }
  if (squot)                    // 単引用文字「'」で囲む場合
  {
    if (quot)
    {
      string tstr;
      tstr = CHRsquot;
      tstr += str;
      tstr += CHRsquot;
      str = tstr;
    }
    else        // 囲まない時だけ文字を特殊表現に変換する
    { strtosstr(str); }
  }
  else                          // 復引用文字「"」で囲む場合
  {
    strtosstr(str);             // 文字を特殊表現に変換してから囲む
    if (quot)
    {
      string tstr;
      tstr = CHRdquot;
      tstr += str;
      tstr += CHRdquot;
      str = tstr;
    }
  }
  return str;
}

int dans::search_data(const set<string>& sset, const string& sdata)
// 文字セット中の文字データをサーチしてインデックスを返す
{
  int sidx = 0;
  bool hit = false;
  set<string>::const_iterator spos = sset.begin();
  while (hit == false && spos != sset.end())
  {
    if (sdata == *spos) hit = true;
    else { ++sidx;  ++spos; }
  }
  return hit ? sidx : -1;       // -1:なし
}

int dans::search_data(const set<double>& fset, double fdata)
// 実数セット中の実数データをサーチしてインデックスを返す
{
  int fidx = 0;
  bool hit = false;
  set<double>::const_iterator fpos = fset.begin();
  while (hit == false && fpos != fset.end())
  {
    if (sgndiff(fdata, *fpos) == 0) hit = true;
    else { ++fidx;  ++fpos; }
  }
  return hit ? fidx : -1;       // -1:なし
}

int dans::search_data(const set<pair<string, double>>& targset, const string& sdata)
// 文字と実数セット中の文字データをサーチして番号を返す
{
  int sno = 1;
  bool hit = false;
  set<pair<string, double>>::const_iterator spos = targset.begin();
  while (hit == false && spos != targset.end())
  {
    if (sdata == spos->first) hit = true;
    else { ++sno;  ++spos; }
  }
  return hit ? sno : -1;        // -1:なし
}

int dans::search_data(const set<pair<string, double>>& targset, double fdata)
// 文字と実数セット中の実数データをサーチして番号を返す
{
  int sno = 1;
  bool hit = false;
  set<pair<string, double>>::const_iterator spos = targset.begin();
  while (hit == false && spos != targset.end())
  {
    if (sgndiff(fdata, spos->second) == 0) hit = true;
    else { ++sno;  ++spos; }
  }
  return hit ? sno : -1;        // -1:なし
}

testtype dans::ctest(double& csqr, int& df, const ftable& ftab, bool ys)
// χ2乗検定
{
  csqr = 0.0;  df = 0;
  if (ftab.nr() == 2 && ftab.nc() == 2 && ys) // Yatesの修正付き2x2のχ2乗検定
  {
    double s = (double)(ftab.f(1, 1) + ftab.f(1, 2));
    double f = (double)(ftab.f(2, 1) + ftab.f(2, 2));
    double m = (double)(ftab.f(1, 1) + ftab.f(2, 1));
    double n = (double)(ftab.f(1, 2) + ftab.f(2, 2));
    double g = s + f;
    double sig2 = (double)(ftab.f(1, 1)) * (double)(ftab.f(2, 2)) - (double)(ftab.f(1, 2)) * (double)(ftab.f(2, 1));
    sig2 -= sgn(sig2) * g * 0.5;
    csqr = sig2 * sig2 / s / f / m / n * g;
    df = 1;
    return testcor;             // 修正有
  }
  double sigc = 0.0;            // 修正無しχ2乗検定
  for (size_t i = 1; i <= ftab.nr(); ++i)
  {
    double sig2 = 0.0;
    for (size_t j = 1; j <= ftab.nc(); ++j)
    {
      if (ftab.f(i, j) > 0)
      {
        double x = (double)(ftab.f(i, j));
        sig2 += x * x / (double)(ftab.fctotal(j));
      }
    }
    sigc += sig2 / (double)(ftab.frtotal(i));
  }
  csqr = (double)(ftab.ftotal()) * (sigc - 1.0);
  df = (int)(ftab.nr() - 1) * (int)(ftab.nc() - 1);
  return testorg;               // 修正無
}

double dans::csqr22(int a, int b, int c, int d) // 2x2のχ2乗値を返す
{
  int m = a + b, n = c + d, s = a + c, f = b + d;
  if (m < 1 || n < 1 || s < 1 || f < 1) return 1.0;
  double bunshi = (double)a * (double)d - (double)b * (double)c;
  double csqr = bunshi / (double)s * bunshi / (double)f * (double)(m + n) / (double)m / (double)n;
  return csqr;
}

testtype dans::utest(double& u, double& vu, double& z, const ftable& ftab, testtype ut)
// Mann-WhitneyのU検定(Wilcoxonの順位和検定)
{
  double total = 0.0, sig3 = 0.0, ta = 0.0;
  for (size_t j = 1; j <= ftab.nc(); ++j)
  {
    double gj = ftab.fctotal(j);
    ta += ftab.f(2, j) * (total + (gj + 1.0) * 0.5);
    sig3 += gj * gj * gj - gj;  total += gj;
  }
  double a = ftab.frtotal(2) * (total + 1.0) * 0.5;
  double k = 1.0 - sig3 / (total * total * total - total);
  double dt = ta - a;
  u = ta - (double)(ftab.frtotal(2) * (ftab.frtotal(2) + 1)) * 0.5;
  vu = ftab.frtotal(1) * k * a / 6.0;
  z = 0.0;
  if (sgn(dt) == 0 || sgn(vu) == 0) return testorg; // 修正無
  double seu = sqrt(vu);
  z = (dt - sgn(dt) * 0.5) / seu;
  if (ut == testcor) return testcor;  // 連続修正有
  if (ut == testorg || (ut == testaut && fabs(z) >= 2.0))
  {
    z = dt / seu;
    return testorg;             // 連続修正無
  }
  return testcor;               // 連続修正有
}

int dans::htest(double& csqr, double& kc, vecdbl& tvec, const ftable& ftab)
// Kruskal-WallisのH検定
{
  double total = 0.0, sig3 = 0.0;
  vecdbl rankvec(ftab.nc() + 1);
  for (size_t j = 1; j <= ftab.nc(); ++j)
  {
    double gj = (double)(ftab.fctotal(j));
    rankvec[j] = total + (gj + 1.0) * 0.5;
    total += gj;  sig3 += gj * gj * gj - gj;
  }
  double sigt2 = 0.0;
  for (size_t i = 1; i <= ftab.nr(); ++i)
  {
    double ti = 0.0;
    for (size_t j = 1; j <= ftab.nc(); ++j)
    {
      if (ftab.f(i, j) > 0) ti += (double)(ftab.f(i, j)) * rankvec[j];
    }
    tvec[i] = ti;  sigt2 += ti * ti / (double)(ftab.frtotal(i));
  }
  tvec[0] = total * (total + 1.0) * 0.5;
  kc = 1.0 - sig3 / (total * total * total - total);
  csqr = 6.0 * (sigt2 / tvec[0] - tvec[0] / total) / kc;
  return (int)(ftab.nr()) - 1;  // 自由度を返す
}

double dans::spear(const ftable& ftab) // Spearmanの順位相関係数計算
{
  double total = 0.0, tr = 0.0;
  vecdbl rrvec(ftab.nr() + 1);
  for (size_t i = 1; i <= ftab.nr(); ++i)
  {
    double f = (double)(ftab.frtotal(i));
    rrvec[i] = total + (f + 1.0) * 0.5;
    tr += f * (f * f - 1.0);  total += f;
  }
  double tc = 0.0, sigr = 0.0;
  total = 0.0;
  for (size_t j = 1; j <= ftab.nc(); ++j)
  {
    double f = (double)(ftab.fctotal(j));
    double rc = total + (f + 1.0) * 0.5;
    tc += f * (f * f - 1.0);
    total += f;
    for (size_t i = 1; i <= ftab.nr(); ++i)
    {
      f = (double)(ftab.f(i, j));
      if (f > 0.0) sigr += f * (rc - rrvec[i]) * (rc - rrvec[i]);
    }
  }
  double rs = total * (total * total - 1.0);
  double tmp = (rs - tr) * (rs - tc);
  const double Teps = DBL_EPSILON * fmax(FDmin, fabs(tmp));
  rs = (fabs(tmp) < Teps) ? 0.0 : (rs - (tr + tc) * 0.5 - 6.0 * sigr) / sqrt(tmp);
  if (sgndiff(1.0, rs * rs) == 0) rs = sgn(rs);
  return rs;                    // Spearmanの順位相関係数を返す
}

double dans::kappa(double& vk, double& vk0, const ftable& ftab, bool weight)
// 一致係数κの計算
{
  vk = vk0 = 0.0;
  double pv = 0.0, po = 0.0, pc = 0.0, g = (double)(ftab.ftotal());
  int nd = (int)(min(ftab.nr(), ftab.nc()));
  for (size_t i = 1; i <= ftab.nr(); ++i)
  {
    double pi = (double)(ftab.frtotal(i)) / g;
    for (size_t j = 1; j <= ftab.nc(); ++j)
    {
      double w = (i == j) ? 1.0 : 0.0;
      if (weight) w = 1.0 - (double)(abs((int)i - (int)j)) / (double)(nd - 1);
      double pj = (double)(ftab.fctotal(j)) / g;
      double pij = (double)(ftab.f(i, j)) / g;
      po += w * pij;  pc += w * pi * pj;
      pv += w * pi * pj * (pi + pj);
    }
  }
  double k = 0.0;
  if (sgndiff(pc, 1.0) != 0)
  {
    vk = po * (1.0 - po) / g / (1.0 - pc) / (1.0 - pc); // 推定用分散
    vk0 = (pc + pc * pc - pv) / g / (1.0 - pc) / (1.0 - pc); // 検定用分散
    k =(po - pc) / (1.0 - pc);
  }
  const double Veps = DBL_EPSILON * fmax(FDmin, vk);
  if (vk < Veps) vk = 0.0;
  if (vk0 < Veps) vk0 = 0.0;
  if (fabs(k) < Veps) k = 0.0;
  return k;                     // κを返す
}

double dans::jttest(const ftable& ftab, double& jt, double& tmax, double& vt)
// Jonckheere-Terpstra検定計算
{
  ftable ftabij(2, ftab.nc());  // U計算用度数表オブジェクト
  double z = 0.0;
  jt = 0.0;
  for (size_t i = 1; i < ftab.nr(); ++i)
  {
    for (size_t k = 1; k <= ftab.nc(); ++k) ftabij.assign(1, k, ftab.f(i, k));
    for (size_t j = i + 1; j <= ftab.nr(); ++j)
    {
      for (size_t k = 1; k <= ftab.nc(); ++k) ftabij.assign(2, k, ftab.f(j, k));
      ftabij.set();
      double u = 0.0, vu = 0.0;
      testtype utype = utest(u, vu, z, ftabij, testorg);
      jt += u;
    }
  }
  double a = 0.0, b = 0.0, c = 0.0;
  tmax = 0.0;
  for (size_t i = 1; i <= ftab.nr(); ++i)
  {
    double ni = (double)(ftab.frtotal(i));
    double nn1 = ni * (ni - 1.0);
    tmax += ni * ni;
    a += nn1 * (2.0 * ni + 5.0);
    b += nn1 * (ni - 2.0);
    c += nn1;
  }
  double st1 = 0.0, st2 = 0.0, st3 = 0.0;
  for (size_t k = 1; k <= ftab.nc(); ++k)
  {
    double tk = (double)(ftab.fctotal(k));
    if (tk > 1.0)
    {
      double tt1 = tk * (tk - 1.0);
      st1 += tt1 * (2.0 * tk + 5.0);
      st2 += tt1 * (tk - 2.0);
      st3 += tt1;
    }
  }
  double nt = (double)(ftab.ftotal());
  double nnt1 = nt * (nt - 1.0);
  tmax = (nt * nt - tmax) / 2.0;
  a = nnt1 * (2.0 * nt + 5.0) - a - st1;
  b = b * st2;
  c = c * st3;
  vt = a / 72.0 + b / 36.0 / nnt1 / (nt - 2.0) + c / 8.0 / nnt1;
  z = (vt > 0.0) ? (jt - tmax * 0.5) / sqrt(vt) : 0.0;
  return z;                    // 検定統計量を返す
}

double dans::ftest(const vecdbl& pvec, const ftable& ftab, bool ys)
// 出現率一様性の検定
{
  double csqr = 0.0, a = ys ? 0.5 : 0.0;
  for (size_t i = 1; i <= ftab.nr(); ++i)
  {
    double f0 = pvec[i] * (double)(ftab.ftotal());
    double d = (double)(ftab.f(i, 1)) - f0;
    d = d - sgn(d) * a;
    csqr += d / f0 * d;
  }
  return csqr;
}

double dans::cctest(double& csqr, double& v, const ftable& ftab) // 累積χ2乗検定
{
  int dc = ftab.nc() - 1;
  double nt = (double)(ftab.ftotal());
  vecdbl uvec(dc);
  csqr = 0.0;
  for (size_t j = 1; j < ftab.nc(); ++j)
  {
    double cj = 0.0;
    for (size_t i = 1; i <= ftab.nr(); ++i) cj += (double)(ftab.f(i, j)) / (double)(ftab.frtotal(i)) * (double)(ftab.f(i, j));
    cj = (cj - (double)(ftab.fctotal(j)) / nt * (double)(ftab.fctotal(j))) / (double)(ftab.fctotal(j)) / (nt - (double)(ftab.fctotal(j)));
    csqr += cj;
    uvec[j - 1] = (double)(ftab.fctotal(j)) / (nt - (double)(ftab.fctotal(j)));
  }
  csqr *= nt * nt;
  double w = 0.0;
  for (int j = 0; j < dc - 1; ++j)
  {
    double uj = uvec[j];
    for (int k = j + 1; k < dc; ++k) w += uj / uvec[k];
  }
  double r = dc / (dc + 2.0 * w);
  csqr *= r;  v = r * (double)(ftab.nr() - 1) * dc;
  return r;                     // 補正係数ρを返す
}

ldbl dans::cctop(ldbl csqr, ldbl v) // χ2乗値・小数自由度→p値変換
{
  int df = (int)v;
  ldbl pr = ctop(csqr, df);
  return pr + (v - (ldbl)df) * (ctop(csqr, df + 1) - pr);
}

int dans::rrtest(double& rr, double& vrr, double& ca, double& ch, const listftab& ftablst)
// リスク比検定
{
  double wlrrm = 0.0, w = 0.0, ct = 0.0;
  int nc = 0;
  for (auto ftabi : ftablst)
  {
    double n11 = (double)(ftabi.f(1, 1)), n12 = (double)(ftabi.f(1, 2));
    double n21 = (double)(ftabi.f(2, 1)), n22 = (double)(ftabi.f(2, 2));
    if (ftabi.f(1, 2) < 1 || ftabi.f(2, 2) < 1)
    { n11 += 0.5;  n12 += 0.5;  n21 += 0.5;  n22 += 0.5; }
    double n1 = n11 + n12, n2 = n21 + n22;
    double rri = n22 / n2 * n1 / n12;
    double lrri = log(rri);
    double wi = 1.0 / (1.0 / n22 - 1.0 / n2 + 1.0 / n12 - 1.0 / n1);
    double wlrri = wi * lrri;
    wlrrm += wlrri;  w += wi;
    ct += wlrri * lrri;  ++nc;
  }
  vrr = 1.0 / w;
  double lrrm = wlrrm / w;
  rr = exp(lrrm);
  ca = w * lrrm * lrrm;
  ch = ct - ca;
  const double CHeps = DBL_EPSILON * fmax(FDmin, ct);
  if (ch < CHeps) ch = 0.0;
  return nc;                    // 計算できた分割表数を返す
}

int dans::lgtest(double& odr, double& vodr, double& ca, double& ch, const listftab& ftablst)
// ロジットによるオッズ比の検定
{
  double wlorm = 0.0, w = 0.0, ct = 0.0;
  for (auto ftabi : ftablst)
  {
    double n11 = (double)(ftabi.f(1, 1)), n12 = (double)(ftabi.f(1, 2));
    double n21 = (double)(ftabi.f(2, 1)), n22 = (double)(ftabi.f(2, 2));
    if (ftabi.f(1, 1) == 0 || ftabi.f(1, 2) == 0 || ftabi.f(2, 1) == 0 || ftabi.f(2, 2) == 0)
    { n11 += 0.5;  n12 += 0.5;  n21 += 0.5;  n22 += 0.5; }
    double odri = n11 / n12 * n22 / n21;
    double lodri = log(odri);
    double wi = 1.0 / (1.0 / n11 + 1.0 / n12 + 1.0 / n21 + 1.0 / n22);
    double wlodri = wi * lodri;
    wlorm += wlodri;  w += wi;
    ct += wlodri * lodri;
  }
  vodr = 1.0 / w;
  double lorm = wlorm / w;
  odr = exp(lorm);
  ca = w * lorm * lorm;
  ch = ct - ca;
  const double CHeps = DBL_EPSILON * fmax(FDmin, ct);
  if (ch < CHeps) ch = 0.0;
  return (int)(ftablst.size()) - 1; // 異質性の検定の自由度を返す
}

int dans::mbtest(double& orm, double& cmm, double& cmr, const listftab& ftablst, testtype ct)
// Mantel-Haenzelの検定+Breslow-Dayの検定
{
  orm = 0.0;
  double ortmp = 0.0;
  for (auto ftabi : ftablst)
  {
    double n = (double)(ftabi.ftotal());
    double n11 = (double)(ftabi.f(1, 1)), n12 = (double)(ftabi.f(1, 2));
    double n21 = (double)(ftabi.f(2, 1)), n22 = (double)(ftabi.f(2, 2));
    if (ftablst.size() > 1 && (ftabi.f(1, 1) == 0 || ftabi.f(1, 2) == 0 || ftabi.f(2, 1) == 0 || ftabi.f(2, 2) == 0))
    { n11 += 0.5;  n12 += 0.5;  n21 += 0.5;  n22 += 0.5;  n += 2.0; }
    orm += n11 / n * n22;  ortmp += n12 / n * n21;
  }
  orm = (sgn(ortmp) == 0) ? 0.0 : orm / ortmp;
  double d = 0.0, v = 0.0, a = orm - 1.0;
  cmr = 0.0;
  for (auto ftabi : ftablst)
  {
    double n = (double)(ftabi.ftotal());
    double n1_ = (double)(ftabi.frtotal(1)), n2_ = (double)(ftabi.frtotal(2));
    double n_1 = (double)(ftabi.fctotal(1)), n_2 = (double)(ftabi.fctotal(2));
    double n11 = (double)(ftabi.f(1, 1)), n12 = (double)(ftabi.f(1, 2));
    double n21 = (double)(ftabi.f(2, 1)), n22 = (double)(ftabi.f(2, 2));
    if (ftablst.size() > 1 && (ftabi.f(1, 1) == 0 || ftabi.f(1, 2) == 0 || ftabi.f(2, 1) == 0 || ftabi.f(2, 2) == 0))
    {
      n11 += 0.5;  n12 += 0.5;  n21 += 0.5;  n22 += 0.5;
      n1_ += 1.0;  n2_ += 1.0;  n_1 += 1.0;  n_2 += 1.0;  n += 2.0;
    }
    d += n11 - n_1 / n * n1_;
    v += n_1 / n * n_2 * n1_ / n * n2_ / (n - 1.0);
    double b = (n_1 + n1_) * orm + n2_ - n_1, c = orm * n_1 * n1_;
    double eorm = (sgn(a) == 0) ? 0.0 : (b + sqrt(b * b - 4.0 * a * c)) / 2.0 / a;
    if (eorm < 0.0 || min(n1_, n_1) < eorm) eorm = (b - sqrt(b * b - 4.0 * a * c)) / 2.0 / a;
    b = n11 - eorm;
    double vorm = (sgn(eorm) == 0) ? 0.0 : 1.0 / eorm;
    vorm += (sgn(n1_ - eorm) == 0) ? 0.0 : 1.0 / (n1_ - eorm);
    vorm += (sgn(n_1 - eorm) == 0) ? 0.0 : 1.0 / (n_1 - eorm);
    vorm += (sgn(n2_ - n_1 + eorm) == 0) ? 0.0 : 1.0 / (n2_ - n_1 + eorm);
    cmr += b * b * vorm;
  }
  cmm = (ct == testcor) ? (d - sgn(d) * 0.5) : d;
  const double Veps = DBL_EPSILON * fmax(FDmin, d);
  cmm = (fabs(v) < Veps) ? 0.0 : cmm * cmm / v;
  if (cmr < Veps) cmr = 0.0;
  return (int)(ftablst.size()) - 1; // Breslow-Dayの検定の自由度を返す
}

int dans::mtest(double& rdm, double& vrdm, double& rrm, double& rra, double& rrb, double& rrc, double& orm, double& ora, double& orb, double& orc, double& cmm, double& cmr, const listftab& ftablst, testtype ct)
// Mantel-Haenzelの検定
{
  rdm = vrdm = rrm = rra = rrb = rrc = orm = ora = orb = orc = 0.0;
  double wd = 0.0, d = 0.0, v = 0.0, cmt = 0.0;
  double vda = 0.0, vdb = 0.0;
  for (auto ftabi : ftablst)
  {
    double n = (double)(ftabi.ftotal());
    double n1_ = (double)(ftabi.frtotal(1)), n2_ = (double)(ftabi.frtotal(2));
    double n_1 = (double)(ftabi.fctotal(1)), n_2 = (double)(ftabi.fctotal(2));
    double n11 = (double)(ftabi.f(1, 1)), n12 = (double)(ftabi.f(1, 2));
    double n21 = (double)(ftabi.f(2, 1)), n22 = (double)(ftabi.f(2, 2));
    double dk = n11 - n_1 / n * n1_;
    double vk = n_1 / n * n_2 * n1_ / n * n2_ / (n - 1.0);
    rdm += (n1_ * n22 - n2_ * n12) / n;
    vda += (n2_ * n2_ * n12 - n1_ * n1_ * n22 + n1_ * n2_ * (n1_ - n2_) / 2.0) / n / n;
    vdb += (n22 * n11 + n12 * n21) / 2.0 / n;
    rra += n1_ / n * n22;  rrb += n2_ / n * n12;
    rrc += (n1_ * n2_ * n_2 - n22 * n12 * n) / n / n;
    ora += n11 / n * n22;  orb += n12 / n * n21;
    orc += (n11 * n22 * (n11 + n21 + 1.0) + n12 * n21 * (n11 + n22 + 1.0)) / n / n;
    d += dk;  v += vk;  wd += n1_ / n * n2_;
    if (fabs(vk) >= FDmin) cmt += dk * dk / vk;
  }
  const double Veps = DBL_EPSILON * fmax(FDmin, v);
  rdm = (fabs(wd) < Veps) ? 0.0 : rdm / wd;   // 調整リスク差
  vrdm = (fabs(wd) < Veps) ? 0.0 : (rdm * vda + vdb) / wd / wd; // 調整リスク差の分散
  rrm = (fabs(rrb) < Veps) ? 0.0 : rra / rrb; // 調整リスク比
  orm = (fabs(orb) < Veps) ? 0.0 : ora / orb; // 調整オッズ比
  cmm = (fabs(v) < Veps) ? 0.0 : d * d / v;
  cmr = cmt - cmm;
  if (cmr < Veps) cmr = 0.0;
  if (ct == testcor)            // 修正有
  {
    d -= sgn(d) * 0.5;
    cmm = (v < Veps) ? 0.0 : d * d / v;
  }
  return (int)(ftablst.size()) - 1; // 異質性の検定の自由度を返す
}

int dans::emtest(double& cmm, double& cmr, double& rd, double& vrd, const listftab& ftablst)
// 拡張Mantel検定
{
  listftab::const_iterator fpos = ftablst.begin();
  int nr = fpos->nr(), nc = fpos->nc();
  vecint ctotalvec(nc);
  for (auto ftabi : ftablst)
  {
    for (int j = 0; j < nc; ++j) ctotalvec[j] += ftabi.fctotal(j + 1);
  }
  vecdbl crankvec(nc);
  int nt = 0, n1 = 0, n2 = 0;
  for (int j = 0; j < nc; ++j)
  {
    double rj = (double)nt + (double)(ctotalvec[j] + 1.0) * 0.5;
    crankvec[j] = rj;  nt += ctotalvec[j];
  }
  double cd = 0.0, v = 0.0, cmt = 0.0, w = 0.0;
  for (auto ftabi : ftablst)
  {
    double cdi = 0.0, cxi = 0.0, cmi = 0.0;
    double ni = (double)(ftabi.ftotal());
    double n1_ = (double)(ftabi.frtotal(1)), n2_ = (double)(ftabi.frtotal(2));
    n1 += ftabi.frtotal(1);  n2 += ftabi.frtotal(2);
    double wi = n1_ * n2_ / ni;
    for (int j = 0; j < nc; ++j)
    {
      double n_j = (double)(ftabi.fctotal(j + 1));
      double r_j = crankvec[j];
      cdi += r_j * ((double)(ftabi.f(2, j + 1)) - n_j * n2_ / ni);
      cxi += r_j * r_j * n_j;
      cmi += r_j * n_j;
    }
    double vi = (cxi - cmi * cmi / ni) * n1_ * n2_ / ni / (ni - 1.0);
    cd += cdi;  v += vi;
    cmt += cdi * cdi / vi;
    w += wi;
  }
  const double Veps = DBL_EPSILON * fmax(FDmin, w);
  cmm = (v < Veps) ? 0.0 : cd * cd / v;
  rd = cd / w;  vrd = v / w / w;
  cmr = cmt - cmm;
  if (cmr < Veps) cmr = 0.0;
  return (int)(ftablst.size()) - 1; // 異質性の検定の自由度を返す
}

int dans::gemtest(double& cmm, double& cmr, const listftab& ftablst)
// 一般化拡張Mantel検定
{
  listftab::const_iterator fpos = ftablst.begin();
  int nr = fpos->nr(), nc = fpos->nc();
  vec<double> dveck(nr - 1), dvec(nr - 1), rveck(nc);
  mat<double> wmatk(nr - 1), wmat(nr - 1);
  double cmt = 0.0;
  for (auto ftabi : ftablst)
  {
    double n = 0.0, nk = (double)(ftabi.ftotal());
    double rkm = 0.0, vkr = 0.0;
    for (int j = 0; j < nc; ++j)
    {
      double nkj = (double)(ftabi.fctotal(j + 1));
      double rkj = n + (nkj + 1.0) * 0.5;
      rveck.assign(j, rkj);
      rkm += rkj * nkj / nk;
      vkr += nkj * rkj * rkj;
      n += nkj;
    }
    vkr = (vkr - nk * rkm * rkm) / nk;
    for (int i = 0; i < nr - 1; ++i)
    {
      double nki = (double)(ftabi.frtotal(i + 1)), di = 0.0;
      for (int j = 0; j < nc; ++j)
      {
        di += rveck(j) * ((double)(ftabi.f(i + 1, j + 1)) - nki * (double)(ftabi.fctotal(j + 1)) / nk);
      }
      dveck.assign(i, di);
      for (int j = 0; j < nr - 1; ++j)
      {
        if (j == i) wmatk.assign(i, j, nki * (nk - nki) / (nk - 1.0) * vkr);
        else wmatk.assign(i, j, -nki * (double)(ftabi.frtotal(j + 1)) / (nk - 1.0) * vkr);
      }
    }
    dvec += dveck;  wmat += wmatk;
    wmatk.inv();
    vec tvec(wmatk * dveck);
    double cmk = inp(dveck, tvec);
    cmt += cmk;
  }
  wmat.inv();
  vec tvec(wmat * dvec);
  cmm = inp(dvec, tvec);
  cmr = cmt - cmm;
  const double Ceps = DBL_EPSILON * fmax(FDmin, cmt);
  if (cmr < Ceps) cmr = 0.0;
  return (nr - 1) * (ftablst.size() - 1); // 異質性の検定の自由度を返す
}

double dans::wtest(const ftable& ftab, double& tp, double& vt, testtype wt)
// Wilcoxonの符号付き順位検定
{
  if (ftab.nr() != 2) return 0.0;
  double total = 0.0, sig3 = 0.0;
  tp = 0.0;
  for (size_t j = 1; j <= ftab.nc(); ++j)
  {
    double gj = (double)(ftab.fctotal(j));
    double rank = total + (gj + 1.0) * 0.5;
    tp += ftab.f(1, j) * rank;
    sig3 += gj * gj * gj - gj;  total += gj;
  }
  double tn = total * (total + 1.0) * 0.5;
  double kk = 2.0 * total * (total + 1.0) * (2.0 * total + 1.0);
  double kc = 1.0 - ((sgn(kk) == 0) ? 0.0 : sig3 / kk);
  vt = kc * kk / 48.0;
  double tmp1 = tp - tn * 0.5, tmp2 = sqrt(vt);
  double z = (sgn(tmp2) == 0) ? 0.0 : (tmp1 - sgn(tmp1) * 0.5) / tmp2;
  if (wt == testcor) return z;  // 連続修正有の正規分布近似値を返す
  if (wt == testorg || (wt == testaut && fabs(z) >= 2.0)) z = tmp1 / tmp2;
  return z;                     // 正規分布近似値を返す
}

ldbl dans::calmcpr(mctype mc, double& f, int g, int df1, int df2, int dc, double r)
// 多重比較の有意確率を計算する(df2==0:ノンパラ)
{
  if (dc < 1) return 1.0;       // 計算できず
  ldbl pr = 0.0L, fldbl = (ldbl)f;
  switch (mc)
  {
  case mctuk : fldbl = sqrt(fldbl * 2.0L);
    pr = qtop(fldbl, g, df2);
    break;
  case mcdun : fldbl = sqrt(fldbl);
    pr = dtop(fldbl, g, df2, r);
    break;
  default : pr = ((df2 > 0) ? ftop(fldbl, df1, df2) : ctop(fldbl, df1)) * dc;
  }
  f = (double)fldbl;
  if (pr < 0.0L) pr = 0.0L;
  return (pr > 1.0L) ? 1.0L : pr; // 有意確率を返す
}

double dans::calmcci(mctype mc, double cc, int g, int df1, int df2, int dc, double vr, double r)
// 多重比較の信頼区間を計算する(df2==0:ノンパラ)
{
  if (dc < 1) return 0.0;       // 計算できず
  ldbl pr = 1.0L - cc, ci = 0.0L;
  switch (mc)
  {
  case mctuk : ci = ptoq(pr, g, df2) / sqrt(2.0L) * sqrt((ldbl)vr);
    break;
  case mcdun : ci = ptod(pr, g, df2, r) * sqrt((ldbl)vr);
    break;
  default : ci = sqrt(((df2 > 0) ? (ldbl)df1 * ptof(pr / (ldbl)dc, df1, df2) : ptoc(pr / (ldbl)dc, df1)) * (ldbl)vr);
  }
  return (double)ci;            // 絶対精度を返す
}

double dans::calbexci(double& plow, double& pup, int n, int nt, double cc)
// F分布を利用した出現率の信頼区間を返す
{
  ldbl ah = (1.0L - cc) * 0.5L; // α/2
  plow = 0.0;
  if (n > 0)                    // 下限
  {
    int df1h = nt - n + 1;
    ldbl f = ptof(ah, 2 * df1h, 2 * n);
    plow = (double)n / ((double)n + (double)df1h * (double)f);
  }
  pup = 1.0;
  if (n < nt)                   // 上限
  {
    int df1h = nt - n, df2h = n + 1;
    ldbl f = ptof(1.0L - ah, 2 * df1h, 2 * df2h);
    pup = (double)df2h / ((double)df2h + (double)df1h * (double)f);
  }
  return pup - plow;            // 信頼区間幅を返す
}

double dans::calbnmci(double& plow, double& pup, double p0, int nt, double cc, bool cor)
// 正規近似を利用した出現率の信頼区間を返す
{
  ldbl pci = ptoz(1.0L - cc) * sqrt((ldbl)(p0 * (1.0 - p0)) / (ldbl)nt);
  if (cor && sgn(pci) != 0) pci += 0.5L / (ldbl)nt;
  plow = p0 - pci;  pup = p0 + pci;
  if (plow < 0.0L) plow = 0.0L;
  if (1.0L < pup) pup = 1.0L;
  return (double)pci;           // 絶対精度を返す
}

double dans::caldpci(double& d, mctype mc, double cc, int g, int df, int dc, int n1, int n2, int m1, int m2, bool cor, double r)
// 出現率の差の信頼区間幅を返す(多重比較含む)
{
  if (dc < 1 || n1 + n2 < 1 || m1 + m2 < 1) return 0.0; // 計算できず
  ldbl nt = (ldbl)(n1 + n2), mt = (ldbl)(m1 + m2);
  d = (double)m2 / mt - (ldbl)n2 / nt;
  ldbl vr = (ldbl)n1 / nt * (ldbl)n2 / nt / nt + (ldbl)m1 / mt * (ldbl)m2 / mt / mt;
  ldbl pr = 1.0L - cc, ci = 0.0L;
  switch (mc)
  {
  case mctuk : ci = ptoq(pr, g, 0) / sqrt(2.0L) * sqrt(vr);
    break;
  case mcdun : ci = ptod(pr, g, 0, r) * sqrt(vr);
    break;
  default : ci = sqrt(ptoc(pr / dc, df) * vr);
  }
  if (cor && sgn(ci) != 0) ci += 0.5L / nt + 0.5L / mt; // 連続修正
  return (double)ci;            // 絶対精度を返す
}

int dans::calfci(double& fl, double& fu, double cc, int df1, int df2, double f)
// 分散比Fの信頼区間を返す
{
  const ldbl Feps = LDBL_EPSILON * fmax((ldbl)FDmin, (ldbl)f); // F計算誤差
  fl = fu = f;
  ldbl pr = (ldbl)((1.0 - cc) * 0.5);
  ldbl flow = ptof(pr, df1, df2), fup = ptof(1.0L - pr, df1, df2);
  int calok = 1;                // 計算結果フラッグ
  if (flow < Feps) calok = -1;    // 下限計算できず
  else fl = f / (double)flow;
  if (fup < Feps) calok += -2;   // 上限計算できず
  else fu = f / (double)fup;
  return calok;     // 1:計算OK -1:下限エラー -2:上限エラー -3:上下限エラー
}

int dans::calcrci(double& crl, double& cru, double cc, int df1, int df2, double f)
// 相関比の信頼区間を返す
{
  const ldbl Feps = LDBL_EPSILON * fmax((ldbl)FDmin, (ldbl)f); // F計算誤差
  crl = cru = sqrt((double)df1 / ((double)df1 + (double)df2 / f));
  ldbl lamda = (ldbl)df1 * (ldbl)f;
  ldbl pr = (ldbl)((1.0 - cc) * 0.5);
  ldbl fl = lptoncf(pr, df1, df2, lamda);
  ldbl fu = lptoncf(1.0L - pr, df1, df2, lamda);
  int calok = 1;                // 計算結果フラッグ
  if (fl < Feps) calok = -1;    // 下限計算できず
  else crl = sqrt((double)df1 / ((double)df1 + (double)df2 / (double)fl));
  if (fu < Feps) calok += -2;   // 上限計算できず
  else cru = sqrt((double)df1 / ((double)df1 + (double)df2 / (double)fu));
  return calok;     // 1:計算OK -1:下限エラー -2:上限エラー -3:上下限エラー
}

int dans::calcsqrci(double& csqrl, double& csqru, double cc, int df, double csqr)
// χ2乗値の信頼区間を返す
{
  csqrl = csqru = csqr;
  ldbl pr = (ldbl)((1.0 - cc) * 0.5);
  csqrl = (double)(lptoncc(pr, df, csqr));
  csqru = (double)(lptoncc(1.0L - pr, df, csqr));
  int calok = 1;                                   // 計算結果フラッグ
  if (csqrl > csqr) calok = -1;  // 下限値非合理
  if (csqru < csqr) calok += -2; // 上限値非合理
  return calok;     // 1:計算OK -1:下限値非合理 -2:上限値非合理 -3:上下限値非合理
}

double dans::calrci(double& zr, double cc, int n, double r)
// 相関係数の信頼区間幅を計算する
{
  if (n < 4 || fabs(r) >= 1.0) return 0.0; // 計算できず
  zr = log((1.0 + r) / (1.0 - r)) * 0.5;
  ldbl pr = 1.0L - cc;
  ldbl ci = ptoz(pr) / sqrt((ldbl)(n - 3));
  return (double)ci;            // 信頼区間幅を返す
}

void dans::calelp(double& a, double& b, double& h, double cc, int df1, int df2, double vxx, double vyy, double vxy)
// 信頼楕円を計算する
{
  a = b = h = 0.0;
  if (df1 < 1 || df2 < 1) return; // 計算できず
  const double Veps = DBL_EPSILON * fmax(FDmin, fmax(vxx, vyy));
  if (vxx < Veps) vxx = 0.0;
  if (vyy < Veps) vyy = 0.0;
  if (fabs(vxy) < Veps) vxy = 0.0;
  if (vxx <= 0.0 && vyy <= 0.0) return;
  ldbl pr = 1.0L - cc;
  ldbl f = ptof(pr, df1, df2);
  double tmp = (vxx + vyy) * (vxx + vyy) - 4.0 * (vxx * vyy - vxy * vxy);
  double root = (tmp < Veps) ? 0.0: sqrt(tmp);
  double t1 = (vxx + vyy - root) * 0.5, t2 = (vxx + vyy + root) * 0.5;
  tmp = (double)df1 * (double)f * (vxx * vyy - vxy * vxy);
  a = (fabs(t1) < Veps) ? 0.0 : sqrt(tmp / t1);
  b = (fabs(t2) < Veps) ? 0.0 : sqrt(tmp / t2);
  h = atan2(2.0 * vxy, vxx - vyy) * 0.5;
}

int dans::setpcvec(vecint& pcvec, const lstr& klst, int pmax)
// 周期成分番号を周期成分番号ベクトル配列にセットする
{
  if (klst.empty()) return 0;
  if (klst.size() > (size_t)pmax) return -1; // 周期成分指定が多すぎる
  pcvec.resize(klst.size());
  size_t m = 0;
  int pcmax = 0;
  for (auto kstr : klst)
  {
    char* ptr = nullptr;
    int k = (int)(strtol(kstr.c_str(), &ptr, 0));
    if (k < 1 || pmax < k) return -2; // 周期成分指定が間違い
    for (size_t i = 0; i < m; ++i)
    {
      if (pcvec[i] == k) return -3; // 同じ周期成分番号がある
    }
    if (k > pcmax) pcmax = k;
    pcvec[m] = k;
    ++m;
  }
  return pcmax;                 // 周期成分番号最大値を返す
}
