// idde.C Version 7.40e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-08 09:42:24 tac>
//
// データ識別情報クラスのメンバー関数
//
// ライブラリ:libdanse.so.7.4
// コンパイル法:g++ -O -fPIC -c idde.C

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const element& element::operator=(const element& element0) // コピー代入演算子
{
  if (&element0 == this) return *this; // 自分自身への代入を防ぐ
  vtag = element0.vtag;  vpcdata = element0.vpcdata;
  vatt = element0.vatt;  vaval = element0.vaval;  vestr = element0.vestr;
  return *this;
}

void element::clear()           // 要素を削除する
{
  vtag = tagnon;  vpcdata.erase();  vatt = attnon;
  vaval = avalnon;  vestr.erase();
}

tagtype element::tag(tagtype tag0) // タグの種類を設定する
{
  if (tag0 == vtag) return vtag; // 変更せず
  vtag = tag0;  vestr.erase();  // 要素の文字列表現をクリアする
  return vtag;
}

const string& element::pcdata(const string& pcdata0) // 内容を設定する
{
  if (pcdata0 == vpcdata) return vpcdata; // 変更せず
  vpcdata = pcdata0;  vestr.erase(); // 要素の文字列表現をクリアする
  return vpcdata;
}

atttype element::att(atttype att0) // 属性を設定する
{
  if (att0 == vatt) return vatt; // 変更せず
  vatt = att0;  vestr.erase();  // 要素の文字列表現をクリアする
  return vatt;
}

avaltype element::aval(avaltype aval0) // 属性値を設定する
{
  if (aval0 == vaval) return vaval; // 変更せず
  vaval = aval0;  vestr.erase(); // 要素の文字列表現をクリアする
  return vaval;
}

const string& element::assign(tagtype tag0, const string& pcdata0, atttype att0, avaltype aval0)
// 要素を設定する
{
  tag(tag0);  pcdata(pcdata0);  att(att0);  aval(aval0);
  return estr();                // 変更後の文字列表現を返す
}

const string& element::estr()   // 要素の文字列表現を返す
{
  if (!(vestr.empty())) return vestr; // 現在の文字列表現を返す
  string tname;
  switch (vtag)
  {
  case ttitle : tname = TAGtitle;  break;
  case tstra : tname = TAGstra;  break;
  case titem : tname = TAGitem;  break;
  }
  string aname = (vatt == atype) ? ATTtype : "";
  string vname;
  switch (vaval)
  {
  case avaland : vname = AVALand;  break;
  case avalor : vname = AVALor;  break;
  case avalrow : vname = AVALrow;  break;
  case avalcol : vname = AVALcol;  break;
  }
  ostringstream ostr;
  ostr << CHRtags << tname;
  if (!aname.empty())
  {
    ostr << " " << aname;
    if (!vname.empty()) ostr << "=" << vname;
  }
  string tstr(vpcdata);
  strtodstr(tstr);              // dans用特殊表現に変換する
  ostr << CHRtage << tstr << CHRtags << CHRetag << tname << CHRtage;
  vestr = ostr.str();
  return vestr;
}

tagtype element::get(const string& str) // 文字列から要素を取得する
{
  if (str.empty() || str[0] != CHRtags) return tagnon; // 文字列不適
  string tstr(str);
  string tagstr;
  tstr.erase(0, 1);             // 最初のCHRtagsを削除する
  strsize_t pos = getsubstr(tagstr, tstr, CHRtage);
  if (pos == string::npos || pos >= tstr.size() || tstr[pos] != CHRtage) return tagnon; // タグ不適
  tstr.erase(0, pos + 1);       // 開始タグ削除
  string substr;
  pos = getsubstr(substr, tagstr, ' '); // タグ名取得
  strtrm(substr);
  tagtype tag = search_tag(substr);
  if (tag == tagetc) return tag; // 不明タグ
  atttype att = attnon;
  avaltype aval = avalnon;
  if (pos != string::npos && pos < tagstr.size() && tagstr[pos] == ' ')
  {                             // 属性指定あり
    tagstr.erase(0, pos + 1);   // タグ名削除
    pos = getsubstr(substr, tagstr, '='); // 属性名取得
    strtrm(substr);
    att = search_att(substr);
    if (att == attetc) return tagetc; // タグの属性不明
    if (pos != string::npos && pos < tagstr.size() && tagstr[pos] == '=')
    {                           // 属性値あり
      tagstr.erase(0, pos + 1); // 属性名削除
      strtrm(tagstr);           // 前後の空白を削除する
      strunquot(tagstr);        // 前後の""を削除する
      aval = search_aval(tagstr);
      if (aval == avaletc) return tagetc;
    }
  }
  pos = getsubstr(substr, tstr, CHRtags); // 内容を取得する
  strtrm(substr);               // 前後の空白を削除する
  sstrtostr(substr);            // 特殊文字を変換する(引用文字も削除)
  vtag = tag;  vpcdata = substr;  vatt = att;  vaval = aval;
  vestr.erase();
  estr();
  return vtag;
}

tagtype element::search_tag(const string& tname) const
// タグ名からタグの種類をサーチして返す
{
  if (tname.empty()) return tagnon;
  if (tname == TAGtitle) return ttitle;
  if (tname == TAGstra) return tstra;
  if (tname == TAGitem) return titem;
  return tagetc;
}

atttype element::search_att(const string& aname) const
// 属性名から属性の種類をサーチして返す
{
  if (aname.empty()) return attnon;
  if (aname == ATTtype) return atype;
  return attetc;
}

avaltype element::search_aval(const string& avname) const
// 属性値名から属性値の種類をサーチして返す
{
  if (avname.empty()) return avalnon;
  if (avname == AVALand) return avaland;
  if (avname == AVALor) return avalor;
  if (avname == AVALrow) return avalrow;
  if (avname == AVALcol) return avalcol;
  return avaletc;
}

const idd& idd::operator=(const idd& idd0) // コピー代入演算子
{
  if (&idd0 == this) return *this; // 自分自身への代入を防ぐ
  vetitle = idd0.vetitle;  vestravec = idd0.vestravec;
  veitemvec = idd0.veitemvec;  edummy = idd0.edummy;
  vnc = idd0.vnc;  iname = idd0.iname;
  return *this;
}

void idd::clear()               // データ識別情報を削除する
{
  vetitle.clear();  vestravec.clear();  veitemvec.clear();
  edummy.clear();  vnc = 0;  iname.erase();
}

size_t idd::set(const string& title, size_t ni0) // データ識別情報を設定する
{
  element ele(ttitle, title);   // タイトルの設定
  vetitle = ele;
  ele.assign(titem, "");        // 項目情報の設定(層別情報は変更しない)
  size_t niold = veitemvec.size();
  veitemvec.resize(ni0, ele);
  for (size_t i = niold; i < ni0; ++i)
  {
    ostringstream ostr;
    ostr << "ITEM" << i + 1;    // 項目名を"item<項目番号>"にする
    veitemvec[i].assign(titem, ostr.str());
  }
  return veitemvec.size();      // 設定後の項目数を返す
}

const string& idd::assign_item(size_t idx, const string& pcdata0, atttype att0, avaltype aval0)
// 項目情報を設定する
{
  iname.erase();
  if (idx >= veitemvec.size()) return iname; // 引数不適
  return veitemvec[idx].assign(titem, pcdata0, att0, aval0);
}

int idd::put(ostream& fout, formtype formt) // データ識別情報をファイルに出力する
{
  if (veitemvec.size() < 1) return 0; // データ識別情報無し
  int nl = put_title(fout, formt);
  nl += put_stra(fout, formt);
  nl += put_item(fout, formt);
  return nl;                    // 出力行数を返す
}

int idd::put_title(ostream& fout, formtype formt) // データ名をファイルに出力する
{
  switch (formt)
  {
  case formtag : fout << vetitle.estr() << "\n"; // タグ形式で出力
    return 1;
  case formngp : fout << "%DATA:" << vetitle.pcdata() << "\n%\n";
    return 2;
  default : fout << "DATA:" << vetitle.pcdata() << "\n\n";
    return 2;
  }
  return 0;                     // 出力行数を返す
}

int idd::put_stra(ostream& fout, formtype formt) // 層別情報をファイルに出力する
{
  if (vestravec.size() < 1) return 0; // 層別情報無し
  switch (formt)
  {
  case formtag :                // タグ形式で出力
    for (size_t i = 0; i < vestravec.size(); ++i)
    { fout << vestravec[i].estr() << "\n"; }
    return vestravec.size();    // 出力行数を返す
  case formngp :                // Ngraph形式で出力
    for (size_t i = 0; i < vestravec.size(); ++i)
    {
      if (vestravec[i].aval() == avaland) fout << "%  AND   :";
      else if (vestravec[i].aval() == avalor) fout << "%  OR    :";
      else fout << "%STRATIFY:";
      fout << vestravec[i].pcdata() << "\n";
    }
    fout << "%\n";
    return vestravec.size() + 1; // 出力行数を返す
  default :                     // 帳票形式で出力
    for (size_t i = 0; i < vestravec.size(); ++i)
    {
      if (vestravec[i].aval() == avaland) fout << "  AND   :";
      else if (vestravec[i].aval() == avalor) fout << "  OR    :";
      else fout << "STRATIFY:";
      fout << vestravec[i].pcdata() << "\n";
    }
    fout << "\n";
    return vestravec.size() + 1; // 出力行数を返す
  }
  return 0;                     // 出力行数を返す
}

int idd::put_item(ostream& fout, formtype formt) // 項目情報をファイルに出力する
{
  for (size_t i = 0; i < veitemvec.size(); ++i)
  {
    switch (formt)
    {
    case formtag : fout << veitemvec[i].estr(); // タグ形式で出力
      break;
    case formngp : fout << "%ITEM" << setw(4) << i + 1 << setw(0);
      fout << ":" << veitemvec[i].pcdata();
      break;
    default : fout << "ITEM" << setw(4) << i + 1 << setw(0);
      fout << ":" << veitemvec[i].pcdata();
    }
    fout << "\n";
  }
  return veitemvec.size();      // 出力行数を返す
}

int idd::put_item(ostream& fout, size_t idx, formtype formt)
// 項目をファイルに出力する
{
  if (veitemvec.size() <= idx) return 0;
  switch (formt)
  {
  case formtag : fout << veitemvec[idx].estr();
    break;
  case formngp : fout << "%ITEM" << setw(4) << idx + 1 << setw(0);
    fout << ":" << veitemvec[idx].pcdata();
    break;
  default : fout << "ITEM" << setw(4) << idx + 1 << setw(0);
    fout << ":" << veitemvec[idx].pcdata();
  }
  fout << "\n";
  return 1;                     // 出力行数を返す
}

int idd::append(const string& str) // データ識別情報を追加する
{
  if (str.empty()) return -1;   // 引数不適
  if (vnc > 0) return -2;       // ケースデータ中に識別情報あり
  element ele;
  switch (ele.get(str))
  {
  case ttitle : vetitle = ele;  // データ名
    break;
  case tstra : vestravec.push_back(ele); // 層別情報
    break;
  case titem : veitemvec.push_back(ele); // 項目情報
    break;
  case tagnon : return -3;      // タグが無い場合はエラーにする
  case tagetc : return -4;      // 不明な場合はエラーにする
  default : return -5;          // その他もエラーにする
  }
  return veitemvec.size();      // 追加後の項目数を返す
}

int idd::append_stra(const string& pcdata0, atttype att0, avaltype aval0)
// 層別情報を追加する
{
  element ele(tstra, pcdata0, att0, aval0);
  vestravec.push_back(ele);
  return vestravec.size();      // 追加後の層別数を返す
}

int idd::append_stra(lstra& stralst) // 多重層別情報を追加する
{
  if (stralst.size() < 1) return 0; // 層別情報無し
  atttype att = attnon;
  avaltype aval = avalnon;
  if (vestravec.size() > 0) { att = atype;  aval = avaland; } // ANDで追加
  int err = 0;
  lstra_itr pos = stralst.begin();
  while (!err && pos != stralst.end())
  {
    if (veitemvec.size() < pos->straino())
    { err = -1; }
    else
    {
      append_stra(pos->strastr(veitemvec[pos->straino() - 1].pcdata()), att, aval);
      att = atype;  aval = avalor;  ++pos; // ORで追加
    }
  }
  return err;
}

int idd::append_item(const string& pcdata0, atttype att0, avaltype aval0)
// 項目情報を追加する
{
  element ele(titem, pcdata0, att0, aval0);
  veitemvec.push_back(ele);
  return veitemvec.size();      // 追加後の項目数を返す
}
