// putstate.C Version 7.40e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-08 09:28:18 tac>
//
// 統計量出力関数群
//
// ライブラリ:libdanse.so.7.4
// コンパイル法:g++ -O -fPIC -c putstate.C

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

int dans::puttitle(ostream& fout, const string& title, int wline)
// DANSのタイトルとバージョンを出力する
{
  int w = strclen(title, CHARwide); // titleの表示桁数
  fout << title << setw(wline - 2 - w) << Version << endl;
  return w;                     // titleの表示桁数を返す
}

double dans::putatab(ostream& fout, const string& factor, double sa, int dfa, double sr, int dfr, double st, int dft)
// 分散分析表を出力して残差分散を返す
{
  constexpr int LENline = 78;   // 区切り線の長さ
  constexpr int LENhed = 10;    // ヘッダの長さ
  if (factor.empty() || st <= 0.0) return 0.0; // 出力なし
  double v = 0.0, f = 0.0, p = 1.0;
  double vr = calvfp(v, f, p, sa, dfa, sr, dfr);
  string tstr("ANOVA Table");
  fout << right << setw((LENline - strclen(tstr, CHARwide)) / 2 + tstr.size()) << tstr << "\n";
  fout << "FACTOR    Sum of Squares        DF  Mean Squares     F-value       p-value\n";
  putline(fout, LENline);
  int npad = factor.size() - strclen(factor, CHARwide);
  fout << left << setw(LENhed + npad) << factor << right;
  fout << setw(14) << sa << setw(10) << dfa << setw(14) << v;
  fout << setw(12) << f << setw(14) << p << pind(p) << "\n";
  tstr = "RESIDUALS";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(LENhed + npad) << tstr << right;
  fout << setw(14) << sr << setw(10) << dfr << setw(14) << vr << "\n";
  putline(fout, LENline);
  tstr = "TOTAL";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(LENhed + npad) << tstr << right;
  fout << setw(14) << st << setw(10) << dft << endl;
  return vr;                    // 残差分散を返す
}

double dans::calvfp(double& v, double& f, double& p, double ss, int df, double sr, int dfr)
// 分散、F値、p値を計算して分散を返す
{
  double vr = (dfr > 0) ? sr / dfr : 0.0;
  v = (df > 0) ? ss / df : 0.0;
  f = (vr > 0.0) ? v / vr : 0.0;
  p = ftop(f, df, dfr);
  return vr;
}

void dans::putmsd(ostream& fout, int n, double mean, double v)
// 例数、平均値、標準偏差を出力する
{
  double sd = 0.0, se = 0.0;
  if (n > 1) { sd = sqrt(v);  se = sqrt(v / (double)n); }
  fout << left << "n=" << setw(13) << n;
  fout << "mean=" << setw(13) << mean;
  fout << "SD=" << setw(13) << sd;
  fout << "SE=" << setw(13) << se;
  fout << right << endl;
}
