// strae.C Version 7.40e  Copyleft (c) 杉本典夫
// Time-stamp: <2023-10-28 10:00:43 tac>
//
// 層別クラスのメンバー関数
//
// ライブラリ:libdanse.so.7.4
// コンパイル法:g++ -O -fPIC -c strae.C

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const stra& stra::operator=(const stra& stra0) // コピー代入演算子
{
  if (&stra0 == this) return *this; // 自分自身への代入を防ぐ
  vstrastr = stra0.vstrastr;  vstratype = stra0.vstratype;
  vstraino = stra0.vstraino;
  vrange.first = stra0.vrange.first;  vrange.second = stra0.vrange.second;
  vstraslst = stra0.vstraslst;
  return *this;
}

void stra::clear()              // 層別情報を削除する
{
  vstrastr.erase();  vstratype = datanon;  vstraino = 0;
  vrange.first = vrange.second = BLANKf;  vstraslst.clear();
}

const string& stra::strastr(const string& iname) // 層別情報の文字列表現を返す
{
  if (!(vstrastr.empty()) || vstratype == datanon) return vstrastr;
  vstrastr = iname + "=";
  if (vstratype == dfdata)      // 実数範囲層別情報
  {
    ostringstream ostr;
    ostr << vrange.first << "-" << vrange.second;
    vstrastr += ostr.str();
  }
  else                          // 文字データ層別情報
  {
    for (lstr_citr pos = vstraslst.begin(); pos != vstraslst.end(); ++pos)
    {
      if (pos != vstraslst.begin()) vstrastr += " OR ";
      string tstr(*pos);
      if (tstr[0] == CHRhead) tstr.erase(0, 1); // 先頭の^を削除する
      if (tstr[tstr.size() - 1] == CHRtail) tstr.resize(tstr.size() - 1); // 末尾の$を削除する
      vstrastr +=  tstr;
    }
  }
  return vstrastr;
}

datatype stra::get(char strach, size_t straino, const string& str)
// 層別情報を取得する
{
  if (straino < 1) return datanon; // 層別情報不適
  vstrastr.erase();
  vstraino = straino;
  if (strach == 'r')            // 実数データ層別情報
  {
    vstratype = dfdata;  vstraslst.clear();
    vrange.first = -DBL_MAX;  vrange.second = DBL_MAX;
    if (getrange(vrange.first, vrange.second, str) != 0) // 実数範囲取得
    {
      if (vrange.first > vrange.second) return datanon; // 実数範囲不適
    }
  }
  else if (strach == 's')       // 文字データ層別情報
  {
    vstratype = dsdata;  vrange.first = vrange.second = BLANKf;
    lstr strlst;
    if (strsplit(strlst, str, CHRdelm) < 1)
    {                           // 文字データがない場合はヌル文字指定と解釈
      string tstr;
      tstr = CHRhead;
      tstr += CHRtail;
      vstraslst.push_back(tstr);
    }
    else
    {
      for (lstr_itr pos = strlst.begin(); pos != strlst.end(); ++pos)
      {
        strtrm(*pos);           // 前後の空白を削除する
        sstrtostr(*pos);        // 特殊表現文字を変換する(引用文字も削除)
        if (pos->size() < 1)
        {
          *pos = CHRhead;
          *pos += CHRtail;
        }
        else
        {
          if ((*pos)[0] != CHRhead)
          {
            string tstr(*pos);
            *pos = CHRhead;
            *pos += tstr;
          }
          if ((*pos)[pos->size() - 1] != CHRtail) *pos += CHRtail;
        }
        vstraslst.push_back(*pos);
      }
    }
  }
  else                          // その他の層別情報
  { return dataetc; }
  return vstratype;             // 層別項目の種別を返す
}

bool stra::match(datac& cdata) const // 層別条件に当てはまるかどうか判定する
{
  if (vstratype == datanon) return true; // 層別情報無し
  bool hit = false;
  if (vstratype == dfdata)      // 実数範囲チェック
  {
    double fdata = cdata.fdata(vstraino - 1);
    if (fdata != BLANKf && vrange.first <= fdata && fdata <= vrange.second) hit = true;
  }
  else                          // 文字データチェック
  {
    lstr_citr spos = vstraslst.begin();
    while (!hit && spos != vstraslst.end()) // ORでチェック
    {
      if (strpmatch(cdata.sdata(vstraino - 1), *spos) > 0) hit = true;
      else ++spos;
    }
  }
  return hit;
}

bool dans::straok(const lstra& stralst, size_t ni)
// 層別情報の整合性をチェックする
{
  if (stralst.empty()) return true;
  bool ok = true;
  lstra_citr spos = stralst.begin();
  while (ok && spos != stralst.end())
  {
    if (spos->straino() > ni) ok = false; // 層別項目番号が範囲外
    else ++spos;
  }
  return ok;
}

bool dans::stramatch(const lstra& stralst, datac& cdata)
// 多重層別に当てはまるかどうか判定する
{
  if (stralst.empty()) return true;
  bool hit = false;
  lstra_citr spos = stralst.begin();
  while (!hit && spos != stralst.end())
  {
    if (spos->match(cdata)) hit = true; // 多重層別はORで調べる
    else ++spos;
  }
  return hit;
}
